"""The script ``add_staffline_symbols.py`` takes as input a CVC-MUSCIMA
(page, writer) index and a corresponding NodeList file
and adds to the NodeList staffline and staff objects."""
import argparse
import logging
import os
import time

import numpy

from skimage.io import imread
from skimage.morphology import watershed
from skimage.filters import gaussian
from typing import Tuple

from mung.dataset import CvcMuscimaDataset
from mung.io import read_nodes_from_file, export_node_list
from mung.node import Node
from mung.utils import compute_connected_components

from mung.constants import InferenceEngineConstants as _CONST


def build_argument_parser():
    parser = argparse.ArgumentParser(description=__doc__, add_help=True,
                                     formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('-i', '--staff_imfile', action='store',
                        help='The image file with the staff-only image.'
                             ' If not given, will use -n, -w and -r'
                             ' to load it from the CVC-MUSCIMA staff removal'
                             ' ground truth.')

    parser.add_argument('-n', '--number', action='store', type=int,
                        help='Number of the CVC-MUSCIMA page (1 - 20)')
    parser.add_argument('-w', '--writer', action='store', type=int,
                        help='Writer of the CVC-MUSCIMA page (1 - 50)')

    parser.add_argument('-r', '--root', action='store',
                        default=os.getenv('CVC_MUSCIMA_ROOT', None),
                        help='Path to CVC-MUSCIMA dataset root. By default, will attempt'
                             ' to read the CVC_MUSCIMA_ROOT env var. If that does not'
                             ' work, the script will fail.')

    parser.add_argument('-a', '--annot', action='store',  # required=True,
                        help='The annotation file for which the staffline and staff'
                             ' Nodes should be added. If not supplied, default'
                             ' document/collection names will be used and Nodes will'
                             ' be numberd from 0 in the output.')
    parser.add_argument('-e', '--export', action='store',
                        help='A filename to which the output NodeList'
                             ' should be saved. If not given, will print to'
                             ' stdout.')
    parser.add_argument('--stafflines_only', action='store_true',
                        help='If set, will only output stafflines, not other symbols.')

    parser.add_argument('-v', '--verbose', action='store_true',
                        help='Turn on INFO messages.')
    parser.add_argument('--debug', action='store_true',
                        help='Turn on DEBUG messages.')

    return parser


def main(args):
    logging.info('Starting main...')
    _start_time = time.clock()

    ########################################################
    # Load gt image.
    logging.info('Loading staffline image.')
    #  - Initialize Dataset. This checks for the root.

    if args.staff_imfile is None:
        cvc_dataset = CvcMuscimaDataset(root=args.root)
        args.staff_imfile = cvc_dataset.imfile(page=args.number,
                                               writer=args.writer,
                                               distortion='ideal',
                                               mode='staff_only')

    # - Load the image.
    gt = (imread(args.staff_imfile, as_grey=True) * 255).astype('uint8')

    # - Cast as binary mask.
    gt[gt > 0] = 1

    ########################################################
    # Locate stafflines in gt image.
    logging.info('Getting staffline connected components.')

    #  - Get connected components in gt image.
    connected_components, labels, bboxes = compute_connected_components(gt)

    #  - Use vertical dimension of CCs to determine which ones belong together
    #    to form stafflines. (Criterion: row overlap.)
    n_rows, n_cols = gt.shape
    intervals = [[] for _ in range(n_rows)]  # For each row: which CCs have pxs on that row?
    for label, (t, l, b, r) in list(bboxes.items()):
        if label == 0:
            continue
        # Ignore very short staffline segments that can easily be artifacts
        # and should not affect the vertical range of the staffline anyway.
        if (r - l) < 8:
            continue
        for row in range(t, b):
            intervals[row].append(label)

    logging.info('Grouping staffline connected components into stafflines.')
    staffline_components = []  # For each staffline, we collect the CCs that it is made of
    _in_staffline = False
    _current_staffline_components = []
    for r_labels in intervals:
        if not _in_staffline:
            # Last row did not contain staffline components.
            if len(r_labels) == 0:
                # No staffline component on current row
                continue
            else:
                _in_staffline = True
                _current_staffline_components += r_labels
        else:
            # Last row contained staffline components.
            if len(r_labels) == 0:
                # Current staffline has no more rows.
                staffline_components.append(set(_current_staffline_components))
                _current_staffline_components = []
                _in_staffline = False
                continue
            else:
                # Current row contains staffline components: the current
                # staffline continues.
                _current_staffline_components += r_labels

    logging.info('No. of stafflines, with component groups: {0}'
                 ''.format(len(staffline_components)))

    # Now: merge the staffline components into one bbox/mask.
    logging.info('Merging staffline components into staffline bboxes and masks.')
    staffline_bboxes = []
    staffline_masks = []
    for sc in sorted(staffline_components,
                     key=lambda c: min([bboxes[cc][0]
                                        for cc in c])):  # Sorted top-down
        st, sl, sb, sr = n_rows, n_cols, 0, 0
        for component in sc:
            t, l, b, r = bboxes[component]
            st, sl, sb, sr = min(t, st), min(l, sl), max(b, sb), max(r, sr)
        _sm = gt[st:sb, sl:sr]
        staffline_bboxes.append((st, sl, sb, sr))
        staffline_masks.append(_sm)

    # Check if n. of stafflines is divisible by 5
    n_stafflines = len(staffline_bboxes)
    logging.info('\tTotal stafflines: {0}'.format(n_stafflines))
    if n_stafflines % 5 != 0:
        import matplotlib.pyplot as plt
        stafllines_mask_image = numpy.zeros(gt.shape)
        for i, (_sb, _sm) in enumerate(zip(staffline_bboxes, staffline_masks)):
            t, l, b, r = _sb
            stafllines_mask_image[t:b, l:r] = min(255, (i * 333) % 255 + 40)
        plt.imshow(stafllines_mask_image, cmap='jet', interpolation='nearest')
        plt.show()
        raise ValueError('No. of stafflines is not divisible by 5!')

    logging.info('Creating staff bboxes and masks.')

    #  - Go top-down and group the stafflines by five to get staves.
    #    (The staffline bboxes are already sorted top-down.)
    staff_bboxes = []
    staff_masks = []

    for i in range(n_stafflines // 5):
        _sbb = staffline_bboxes[5 * i:5 * (i + 1)]
        _st = min([bb[0] for bb in _sbb])
        _sl = min([bb[1] for bb in _sbb])
        _sb = max([bb[2] for bb in _sbb])
        _sr = max([bb[3] for bb in _sbb])
        staff_bboxes.append((_st, _sl, _sb, _sr))
        staff_masks.append(gt[_st:_sb, _sl:_sr])

    logging.info('Total staffs: {0}'.format(len(staff_bboxes)))

    ##################################################################
    # (Optionally fill in missing pixels, based on full image.)
    logging.info('SKIP: fill in missing pixels based on full image.')
    #  - Load full image
    #  - Find gap regions
    #  - Obtain gap region masks from full image
    #  - Add gap region mask to staffline mask.

    # Create the Nodes for stafflines and staffs:
    #  - Load corresponding annotation, to which the stafflines and
    #    staves should be added. (This is needed to correctly set docname
    #    and node_ids.)
    if not args.annot:
        nodes = []
        next_node_id = 0
        dataset = 'FCNOMR'
        document = os.path.splitext(os.path.basename(args.staff_imfile))[0]
    else:
        if not os.path.isfile(args.annot):
            raise ValueError('Annotation file {0} does not exist!'.format(args.annot))

        logging.info('Creating Nodes...')
        nodes = read_nodes_from_file(args.annot)
        logging.info('Non-staffline Nodes: {0}'.format(len(nodes)))
        next_node_id = max([c.id for c in nodes]) + 1
        dataset = nodes[0].dataset
        document = nodes[0].document

    #  - Create the staffline Nodes
    stafflines = []
    for sl_bb, sl_m in zip(staffline_bboxes, staffline_masks):
        t, l, b, r = sl_bb
        c = Node(id_=next_node_id,
                 class_name=_CONST.STAFFLINE_CLASS_NAME,
                 top=t, left=l, height=b - t, width=r - l,
                 mask=sl_m,
                 dataset=dataset, document=document)
        stafflines.append(c)
        next_node_id += 1

    if not args.stafflines_only:

        #  - Create the staff Nodes
        staffs = []
        for s_bb, s_m in zip(staff_bboxes, staff_masks):
            t, l, b, r = s_bb
            c = Node(id_=next_node_id,
                     class_name=_CONST.STAFF_CLASS_NAME,
                     top=t, left=l, height=b - t, width=r - l,
                     mask=s_m,
                     dataset=dataset, document=document)
            staffs.append(c)
            next_node_id += 1

        #  - Add the inlinks/outlinks
        for i, sc in enumerate(staffs):
            sl_from = 5 * i
            sl_to = 5 * (i + 1)
            for sl in stafflines[sl_from:sl_to]:
                sl.inlinks.append(sc.id)
                sc.outlinks.append(sl.id)

        # Add the staffspaces.
        staffspaces = []
        for i, staff in enumerate(staffs):
            current_stafflines = [sc for sc in stafflines if sc.id in staff.outlinks]
            sorted_stafflines = sorted(current_stafflines, key=lambda x: x.top)

            current_staffspaces = []

            # Percussion single-line staves do not have staffspaces.
            if len(sorted_stafflines) == 1:
                continue

            # Internal staffspace
            for s1, s2 in zip(sorted_stafflines[:-1], sorted_stafflines[1:]):
                # s1 is the UPPER staffline, s2 is the LOWER staffline
                # Left and right limits: to simplify things, we take the column
                # *intersection* of (s1, s2). This gives the invariant that
                # the staffspace is limited from top and bottom in each of its columns.
                l = max(s1.left, s2.left)
                r = min(s1.right, s2.right)

                # Shift s1, s2 to the right by this much to have the cols. align
                # All of these are non-negative.
                dl1, dl2 = l - s1.left, l - s2.left
                dr1, dr2 = s1.right - r, s2.right - r

                # The stafflines are not necessarily straight,
                # so top is given for the *topmost bottom edge* of the top staffline + 1

                # First create mask
                canvas = numpy.zeros((s2.bottom - s1.top, r - l), dtype='uint8')

                # Paste masks into canvas.
                # This assumes that the top of the bottom staffline is below
                # the top of the top staffline... and that the bottom
                # of the top staffline is above the bottom of the bottom
                # staffline. This may not hold in very weird situations,
                # but it's good for now.
                logging.debug(s1.bounding_box, s1.mask.shape)
                logging.debug(s2.bounding_box, s2.mask.shape)
                logging.debug(canvas.shape)
                logging.debug('l={0}, dl1={1}, dl2={2}, r={3}, dr1={4}, dr2={5}'
                              ''.format(l, dl1, dl2, r, dr1, dr2))
                # canvas[:s1.height, :] += s1.mask[:, dl1:s1.width-dr1]
                # canvas[-s2.height:, :] += s2.mask[:, dl2:s2.width-dr2]

                # We have to deal with staffline interruptions.
                # One way to do this
                # is watershed fill: put markers along the bottom and top
                # edge, use mask * 10000 as elevation

                s1_above, s1_below = staffline_surroundings_mask(s1)
                s2_above, s2_below = staffline_surroundings_mask(s2)

                # Get bounding boxes of the individual stafflines' masks
                # that intersect with the staffspace bounding box, in terms
                # of the staffline bounding box.
                s1_t, s1_l, s1_b, s1_r = 0, dl1, \
                                         s1.height, s1.width - dr1
                s1_h, s1_w = s1_b - s1_t, s1_r - s1_l
                s2_t, s2_l, s2_b, s2_r = canvas.shape[0] - s2.height, dl2, \
                                         canvas.shape[0], s2.width - dr2
                s2_h, s2_w = s2_b - s2_t, s2_r - s2_l

                logging.debug(s1_t, s1_l, s1_b, s1_r, (s1_h, s1_w))

                # We now take the intersection of s1_below and s2_above.
                # If there is empty space in the middle, we fill it in.
                staffspace_mask = numpy.ones(canvas.shape)
                staffspace_mask[s1_t:s1_b, :] -= (1 - s1_below[:, dl1:s1.width - dr1])
                staffspace_mask[s2_t:s2_b, :] -= (1 - s2_above[:, dl2:s2.width - dr2])

                ss_top = s1.top
                ss_bottom = s2.bottom
                ss_left = l
                ss_right = r

                staffspace = Node(next_node_id, _CONST.STAFFSPACE_CLASS_NAME,
                                  top=ss_top, left=ss_left,
                                  height=ss_bottom - ss_top,
                                  width=ss_right - ss_left,
                                  mask=staffspace_mask,
                                  dataset=dataset, document=document)

                staffspace.inlinks.append(staff.id)
                staff.outlinks.append(staffspace.id)

                current_staffspaces.append(staffspace)

                next_node_id += 1

            # Add top and bottom staffspace.
            # These outer staffspaces will have the width
            # of their bottom neighbor, and height derived
            # from its mask columns.
            # This is quite approximate, but it should do.

            # Upper staffspace
            tsl = sorted_stafflines[0]
            tsl_heights = tsl.mask.sum(axis=0)
            tss = current_staffspaces[0]
            tss_heights = tss.mask.sum(axis=0)

            uss_top = max(0, tss.top - max(tss_heights))
            uss_left = tss.left
            uss_width = tss.width
            # We use 1.5, so that large noteheads
            # do not "hang out" of the staffspace.
            uss_height = int(tss.height / 1.2)
            # Shift because of height downscaling:
            uss_top += tss.height - uss_height
            uss_mask = tss.mask[:uss_height, :] * 1

            staffspace = Node(next_node_id, _CONST.STAFFSPACE_CLASS_NAME,
                              top=uss_top, left=uss_left,
                              height=uss_height,
                              width=uss_width,
                              mask=uss_mask,
                              dataset=dataset, document=document)
            current_staffspaces.append(staffspace)
            staff.outlinks.append(staffspace.id)
            staffspace.inlinks.append(staff.id)
            next_node_id += 1

            # Lower staffspace
            bss = current_staffspaces[-1]
            bss_heights = bss.mask.sum(axis=0)
            bsl = sorted_stafflines[-1]
            bsl_heights = bsl.mask.sum(axis=0)

            lss_top = bss.bottom  # + max(bsl_heights)
            lss_left = bss.left
            lss_width = bss.width
            lss_height = int(bss.height / 1.2)
            lss_mask = bss.mask[:lss_height, :] * 1

            staffspace = Node(next_node_id, _CONST.STAFFSPACE_CLASS_NAME,
                              top=lss_top, left=lss_left,
                              height=lss_height,
                              width=lss_width,
                              mask=lss_mask,
                              dataset=dataset, document=document)
            current_staffspaces.append(staffspace)
            staff.outlinks.append(staffspace.id)
            staffspace.inlinks.append(staff.id)
            next_node_id += 1

            # ################ End of dealing with upper/lower staffspace ######

            # Add to current list
            staffspaces += current_staffspaces

        # - Join the lists together
        nodes_with_staffs = nodes \
                            + stafflines \
                            + staffspaces \
                            + staffs

    else:
        nodes_with_staffs = nodes + stafflines

    logging.info('Exporting the new Node list: {0} objects'
                 ''.format(len(nodes_with_staffs)))
    # - Export the combined list.
    nodes_string = export_node_list(nodes_with_staffs)
    if args.export is not None:
        with open(args.export, 'w') as hdl:
            hdl.write(nodes_string)
    else:
        print(nodes_string)

    _end_time = time.clock()
    logging.info('add_staffline_symbols.py done in {0:.3f} s'
                 ''.format(_end_time - _start_time))


def staffline_surroundings_mask(staffline: Node) -> Tuple[numpy.ndarray, numpy.ndarray]:
    """Find the parts of the staffline's bounding box which lie
    above or below the actual staffline.

    These areas will be very small for straight stafflines,
    but might be considerable when staffline curvature grows.
    """
    # We segment both masks into "above staffline" and "below staffline"
    # areas.
    elevation = staffline.mask * 255
    # Blur, to plug small holes somewhat:
    elevation = gaussian(elevation, sigma=1.0)
    # Prepare the segmentation markers: 1 is ABOVE, 2 is BELOW
    markers = numpy.zeros(staffline.mask.shape)
    markers[0, :] = 1
    markers[-1, :] = 2
    markers[staffline.mask != 0] = 0
    seg = watershed(elevation, markers)

    bmask = numpy.ones(seg.shape)
    bmask[seg != 2] = 0
    tmask = numpy.ones(seg.shape)
    tmask[seg != 1] = 0

    return bmask, tmask


if __name__ == '__main__':
    parser = build_argument_parser()
    args = parser.parse_args()

    if args.verbose:
        logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.INFO)
    if args.debug:
        logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.DEBUG)

    main(args)



import logging

import numpy
from skimage.measure import label
from typing import Tuple, Dict, Any, List

from mung.constants import InferenceEngineConstants as _CONST


def connected_components2bboxes(labels: List[List[int]]) -> Dict[int, List[int]]:
    """Returns a dictionary of bounding boxes (upper left c., lower right c.)
    for each label.

    >>> labels = [[0, 0, 1, 1], [2, 0, 0, 1], [2, 0, 0, 0], [0, 0, 3, 3]]
    >>> bboxes = connected_components2bboxes(labels)
    >>> bboxes[0]
    [0, 0, 4, 4]
    >>> bboxes[1]
    [0, 2, 2, 4]
    >>> bboxes[2]
    [1, 0, 3, 1]
    >>> bboxes[3]
    [3, 2, 4, 4]


    :param labels: The output of cv2.connectedComponents().

    :returns: A dict indexed by labels. The values are quadruplets
        (xmin, ymin, xmax, ymax) so that the component with the given label
        lies exactly within labels[xmin:xmax, ymin:ymax].
    """
    bboxes = {}
    for x, row in enumerate(labels):
        for y, l in enumerate(row):
            if l not in bboxes:
                bboxes[l] = [x, y, x + 1, y + 1]
            else:
                box = bboxes[l]
                if x < box[0]:
                    box[0] = x
                elif x + 1 > box[2]:
                    box[2] = x + 1
                if y < box[1]:
                    box[1] = y
                elif y + 1 > box[3]:
                    box[3] = y + 1
    return bboxes


def compute_connected_components(image: numpy.ndarray) -> \
        Tuple[int, numpy.ndarray, Dict[int, List[int]]]:
    labels = label(image, background=0)
    number_of_connected_components = int(labels.max())
    bboxes = connected_components2bboxes(labels)
    return number_of_connected_components, labels, bboxes




import os
import unittest

from mung.io import read_nodes_from_file, export_node_list
from mung.node import Node


class NodeTest(unittest.TestCase):
    def test_bbox_to_integer_bounds(self):
        # Arrange
        expected = (44, 18, 56, 93)
        expected2 = (44, 18, 56, 93)

        # Act
        actual = Node.round_bounding_box_to_integer(44.2, 18.9, 55.1, 92.99)
        actual2 = Node.round_bounding_box_to_integer(44, 18, 56, 92.99)

        # Assert
        self.assertEqual(actual, expected)
        self.assertEqual(actual2, expected2)

    def test_overlaps(self):
        # Arrange
        node = Node(0, 'test', 10, 100, height=20, width=10)

        # Act and Assert
        self.assertEqual(node.bounding_box, (10, 100, 30, 110))

        self.assertTrue(node.overlaps((10, 100, 30, 110)))  # Exact match

        self.assertFalse(node.overlaps((0, 100, 8, 110)))  # Row mismatch
        self.assertFalse(node.overlaps((10, 0, 30, 89)))  # Column mismatch
        self.assertFalse(node.overlaps((0, 0, 8, 89)))  # Total mismatch

        self.assertTrue(node.overlaps((9, 99, 31, 111)))  # Encompasses Node
        self.assertTrue(node.overlaps((11, 101, 29, 109)))  # Within Node
        self.assertTrue(node.overlaps((9, 101, 31, 109)))  # Encompass horz., within vert.
        self.assertTrue(node.overlaps((11, 99, 29, 111)))  # Encompasses vert., within horz.
        self.assertTrue(node.overlaps((11, 101, 31, 111)))  # Corner within: top left
        self.assertTrue(node.overlaps((11, 99, 31, 109)))  # Corner within: top right
        self.assertTrue(node.overlaps((9, 101, 29, 111)))  # Corner within: bottom left
        self.assertTrue(node.overlaps((9, 99, 29, 109)))  # Corner within: bottom right

    def test_read_nodes_from_file(self):
        test_data_dir = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'test', 'test_data')
        clfile = os.path.join(test_data_dir, '01_basic.xml')
        nodes = read_nodes_from_file(clfile)
        self.assertEqual(len(nodes), 48)

    def test_read_nodes_from_file_with_data(self):
        test_data_dir = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'test', 'test_data')
        file = os.path.join(test_data_dir, '01_basic_binary.xml')
        nodes = read_nodes_from_file(file)
        self.assertEqual("G", nodes[0].data['pitch_step'])
        self.assertEqual(79, nodes[0].data['midi_pitch_code'])
        self.assertEqual([8, 17], nodes[0].data['precedence_outlinks'])


if __name__ == '__main__':
    unittest.main()



"""This module implements a Grammar.

A Grammar is a set of rules about how objects from a certain set
of classes are allowed to form relationships. In a dependency grammar,
the relationships are formed directly between the objects. (In
constituency grammars, we'd have a "merge result" object instead.)

In the ``mung`` package, you can use grammars to validate
whether the relationships between the annotated objects conform
to the specification.

.. warning::

    The grammar is not a formal specification. Music notation sometimes
    breaks its own rules. More importantly, people who write music
    notation by hand make mistakes. This means that not all annotation
    files will pass grammar validation without errors, and that is fine.
    If this bothers you, use the MUSCIMarker tool to visualize the errors.


.. TODO::
    create image:: ../doc/_static/grammar_explainer.png

"""
import codecs
import logging
import os
import pprint

import collections

from typing import Tuple, List, Set, Dict


class DependencyGrammarParseError(ValueError):
    pass


class DependencyGrammar(object):
    """The DependencyGrammar class implements rules about valid graphs above
    objects from a set of recognized classes.

    The Grammar complements a Parser. It defines rules, and the Parser
    implements algorithms to apply these rules to some input.

    A grammar has an **Alphabet** and **Rules**. The alphabet is a list
    of symbols that the grammar recognizes. Rules are constraints on
    the structures that can be induced among these symbols.

    There are two kinds of grammars according to what kinds of rules
    they use: **dependency** rules, and **constituency** rules. We use
    dependency grammars. Dependency grammar rules specify which symbols
    are governing, and which symbols are governed::

      noteheadFull | stem

    There can be multiple left-hand side and right-hand side symbols,
    as a shortcut for a list of rules::

        noteheadFull | stem beam
        noteheadFull noteheadHalf | legerLine durationDot tie notehead*Small

    The asterisk works as a wildcard. Currently, only one wildcard per symbol
    is allowed::

      timeSignature | numeral*

    Lines starting with a ``#`` are regarded as comments and ignored.
    Empty lines are also ignored.

    **Cardinality rules**

    We can also specify in the grammar the minimum and/or maximum number
    of relationships, both inlinks and outlinks, that an object can form
    with other objects of given types. For example:

    * One notehead may have up to two stems attached.
    * We also allow for stemless full noteheads.
    * One stem can be attached to multiple noteheads, but at least one.

    This would be expressed as::

      notehead*{,2} | stem{1,}

    The relationship of noteheads to leger lines is generally ``m:n``::

      noteheadFull | legerLine

    A time signature may consist of multiple numerals, but only one
    other symbol::

      timeSignature{1,} | numeral*{1}
      timeSignature{1} | timeSigCommon timeSigCutCommon

    A key signature may have any number of sharps and flats.
    A sharp or flat can only belong to one key signature. However,
    not every sharp belongs to a key signature::

      keySignature | accidentalSharp{,1} accidentalFlat{,1} accidentalNatural{,1} accidentalDoubleSharp{,1} accidentalDoubleFlat{,1}

    For the left-hand side of the rule, the cardinality restrictions apply to
    outlinks towards symbols of classes on the right-hand side of the rule.
    For the right-hand side, the cardinality restrictions apply to inlinks
    from symbols of left-hand side classes.

    It is also possible to specify that regardless of where outlinks
    lead, a symbol should always have at least some::

      timeSignature{1,} |
      repeat{2,} |

    And analogously for inlinks::

      | letter*{1,}
      | numeral*{1,}
      | legerLine{1,}
      | noteheadFullSmall{1,}

    **Interface**

    The basic role of the dependency grammar is to provide the list of rules:

    >>> from mung.io import parse_node_classes
    >>> filepath = os.path.dirname(os.path.dirname(__file__)) + u'/test/test_data/mff-muscima-classes-annot.deprules'
    >>> node_classes_path = os.path.dirname(os.path.dirname(__file__)) + u'/test/test_data/mff-muscima-classes-annot.xml'
    >>> node_classes = parse_node_classes(node_classes_path)
    >>> node_classes_dict = {node_class.name for node_class in node_classes}
    >>> dependency_graph = DependencyGrammar(grammar_filename=filepath, alphabet=node_classes_dict)
    >>> len(dependency_graph.rules)
    646

    The grammar can validate against these rules. The workhorse of this
    functionality is the ``find_invalid_in_graph()`` method, which finds
    objects that have inlinks/outlinks which do not comply with the grammar,
    and the non-compliant inlinks/outlinks as well.

    If we have the following notation objects ``0``, ``1``, ``2``, and ``3``,
    with the following symbol classes:

    >>> vertices = {0: 'noteheadFull', 1: 'stem', 2: 'flag8thUp', 3: 'noteheadHalf'}

    And the following relationships were recorded:

    >>> edges = [(0, 1), (0, 2), (0, 3)]

    We can check for errors against our music notation symbols dependency
    grammar:

    >>> wrong_vertices, wrong_inlinks, wrong_outlinks, _, _, _ = \
            dependency_graph.find_invalid_in_graph(vertices=vertices, edges=edges)

    Because the edge ``(0, 3)`` connects a full notehead to an empty notehead,
    the method should report the objects ``0`` and ``3`` as wrong, as well
    as the corresponding inlink of ``3`` and outlink of ``0``:

    >>> wrong_vertices
    [0, 3]
    >>> wrong_inlinks
    [(0, 3)]
    >>> wrong_outlinks
    [(0, 3)]

    (Note that both the inlinks and outlinks are recorded in a ``(from, to)``
    format.)

    .. caution::

        Aside from checking against illegal relationships (such as we
        saw in the example), errors can also come from too many or too
        few inlinks/outlinks of a given type. However,
        the validation currently implements checks only for aggregate
        cardinalities, not for pair cardinalities (so, there can be
        e.g. multiple sharps attached to a notehead, even though the cardinality
        in the ``notehead | sharp`` rule is set to max. 1).

    **Grammar file formats**

    The alphabet is stored by means of a ``NodeClassList`` XML file with
    :class:`NodeClass` elements, as described in the :mod:`mung.io` module.

    The rules are stored in *rule files*, with the suffix ``.deprules``.

    A rule file line can be empty, start with a ``#`` (comment), or contain
    a rule symbol ``|``. Empty lines and comments are ignored during parsing.
    Rules are split into left- and right-hand side tokens, according to
    the position of the ``|`` symbol.

    Parsing a token returns the token string (unexpanded wildcards), its
    minimum and maximum cardinality in the rule (defaults are ``(0, 10000)``
    if no cardinality is provided).

    >>> dependency_graph.parse_token('notehead*')
    ('notehead*', 0, 10000)
    >>> dependency_graph.parse_token('notehead*{1,5}')
    ('notehead*', 1, 5)
    >>> dependency_graph.parse_token('notehead*{1,}')
    ('notehead*', 1, 10000)
    >>> dependency_graph.parse_token('notehead*{,5}')
    ('notehead*', 0, 5)
    >>> dependency_graph.parse_token('notehead*{1}')
    ('notehead*', 1, 1)

    The wildcards are expanded at the level of a line.

    >>> l = 'notehead*{,2} | stem'
    >>> rules, inlink_cards, outlink_cards, _, _ = dependency_graph.parse_dependency_grammar_line(l)
    >>> rules
    [('noteheadFull', 'stem'), ('noteheadFullSmall', 'stem'), ('noteheadHalf', 'stem'), ('noteheadHalfSmall', 'stem'), ('noteheadWhole', 'stem')]
    >>> outlink_cards['noteheadHalf'] == {'stem': (0, 2)}
    True
    >>> inlink_cards['stem'] == {'noteheadHalf': (0, 10000), 'noteheadFull': (0, 10000),
    ...    'noteheadWhole': (0, 10000), 'noteheadFullSmall': (0, 10000), 'noteheadHalfSmall': (0, 10000)}
    True


    A key signature can have any number of sharps, flats, or naturals,
    but if a given symbol is part of a key signature, it can only be part of one.

    >>> l = 'keySignature | accidentalSharp{1} accidentalFlat{1} accidentalNatural{1}'
    >>> rules, inlink_cards, _, _, _ = dependency_graph.parse_dependency_grammar_line(l)
    >>> rules
    [('keySignature', 'accidentalFlat'), ('keySignature', 'accidentalNatural'), ('keySignature', 'accidentalSharp')]
    >>> inlink_cards == {'accidentalNatural': {'keySignature': (1, 1)},
    ...                  'accidentalSharp': {'keySignature': (1, 1)},
    ...                  'accidentalFlat': {'keySignature': (1, 1)}}
    True


    You can also give *aggregate* cardinality rules, of the style "whatever rule
    applies, there should be at least X/at most Y edges for this type of object".
    (If no maximum is specified, the value of ``DependencyGrammar._MAX_CARD``
    is used, which is by default 10000).

    >>> l = 'keySignature{1,} |'
    >>> _, _, _, _, out_aggregate_cards = dependency_graph.parse_dependency_grammar_line(l)
    >>> out_aggregate_cards == {'keySignature': (1, 10000)}
    True
    >>> l = 'notehead*Small{1,} |'
    >>> _, _, _, _, out_aggregate_cards = dependency_graph.parse_dependency_grammar_line(l)
    >>> out_aggregate_cards == {'noteheadFullSmall': (1, 10000), 'noteheadHalfSmall': (1, 10000)}
    True
    >>> l = '| beam{1,} stem{1,} accidentalFlat{1,}'
    >>> _, _, _, in_aggregate_cards, _ = dependency_graph.parse_dependency_grammar_line(l)
    >>> in_aggregate_cards == {'stem': (1, 10000), 'beam': (1, 10000), 'accidentalFlat': (1, 10000)}
    True

    """

    WILDCARD = '*'

    _MAX_CARDINALITY = 10000

    def __init__(self, grammar_filename: str, alphabet: Set[str]):
        """Initialize the Grammar: fill in alphabet and parse rules.

        :param grammar_filename: Path to a file that contains deprules
            (see class documentation for ``*.deprules`` file format).

        :param alphabet: A set or list of symbol class names, which
            are used in the *.deprules file.
        """
        self.alphabet = set(alphabet)
        # logging.info('DependencyGrammar: got alphabet:\n{0}'
        #              ''.format(pprint.pformat(self.alphabet)))
        self.rules = []  # type: List[Tuple[str, str]]
        self.inlink_cardinalities = {}  # type: Dict[str, Dict[str, Tuple[int, int]]]
        self.outlink_cardinalities = {}  # type: Dict[str, Dict[str, Tuple[int, int]]]
        self.inlink_aggregated_cardinalities = {}  # type: Dict[str, Tuple[int, int]]
        self.outlink_aggregated_cardinalities = {}  # type: Dict[str, Tuple[int, int]]

        rules, inlink_cardinalities, outlink_cardinalities, inlink_aggregated_cardinalitites, \
        outlink_aggregated_cardinalitites = self.parse_dependency_grammar_rules(grammar_filename)

        if self.__validate_rules(rules):
            self.rules = rules
            logging.info('DependencyGrammar: Imported {0} rules'
                         ''.format(len(self.rules)))
            self.inlink_cardinalities = inlink_cardinalities
            self.outlink_cardinalities = outlink_cardinalities
            self.inlink_aggregated_cardinalities = inlink_aggregated_cardinalitites
            self.outlink_aggregated_cardinalities = outlink_aggregated_cardinalitites
            logging.debug('DependencyGrammar: Inlink aggregated cardinalities: {0}'
                          ''.format(pprint.pformat(inlink_aggregated_cardinalitites)))
            logging.debug('DependencyGrammar: Outlink aggregated cardinalities: {0}'
                          ''.format(pprint.pformat(outlink_aggregated_cardinalitites)))
        else:
            raise DependencyGrammarParseError(
                'Not able to parse dependency grammar file {0}.'
                ''.format(grammar_filename))

    def validate_edge(self, head_name: str, child_name: str) -> bool:
        """Check whether a given ``head --> child`` edge conforms
        with this grammar."""
        return (head_name, child_name) in self.rules

    def validate_graph(self, vertices: Dict[int, str], edges: List[Tuple[int, int]]):
        """Checks whether the given graph complies with the grammar.

        :param vertices: A dict with any keys and values corresponding
            to the alphabet of the grammar.

        :param edges: A list of ``(from, to)`` pairs, where both
            ``from`` and ``to`` are valid keys into the ``vertices`` dict.

        :returns: ``True`` if the graph is valid, ``False`` otherwise.
        """
        v, i, o, _, _, _ = self.find_invalid_in_graph(vertices=vertices, edges=edges)
        return len(v) == 0

    def find_invalid_in_graph(self, vertices: Dict[int, str], edges: List[Tuple[int, int]]) \
            -> Tuple[List[int], List[Tuple[int, int]], List[Tuple[int, int]], Dict[int, str], Dict[
                Tuple[int, int], str], Dict[Tuple[int, int], str]]:
        """Finds vertices and edges where the given object graph does
        not comply with the grammar.

        Wrong vertices are any that:

        * are not in the alphabet;
        * have a wrong inlink or outlink;
        * have missing outlinks or inlinks.

        Discovering missing edges is difficult, because the grammar
        defines cardinalities on a per-rule basis and there is currently
        no way to make a rule compulsory, or to require at least one rule
        from a group to apply. It is currently not implemented.

        Wrong outlinks are such that:

        * connect symbol pairs that should not be connected based on their
          classes;
        * connect so that they exceed the allowed number of outlinks to
          the given symbol type

        Wrong inlinks are such that:

        * connect symbol pairs that should not be connected based on their
          classes;
        * connect so that they exceed the allowed number of inlinks
          to the given symbol based on the originating symbols' classes.

        :param vertices: A dict with any keys, and values corresponding
            to the alphabet of the grammar.

        :param edges: A list of ``(from, to)`` pairs, where both
            ``from`` and ``to`` are valid keys into the ``vertices`` dict.

        :param provide_reasons: If set, will generate string descriptions
            of each error and return them.

        :returns: A list of vertices, a list of inlinks and a list of outlinks
            that do not comply with the grammar as well as three dictionaries with reasons
            for each list respectively.
        """
        logging.info('DependencyGrammar: looking for errors.')

        wrong_vertices = []  # type: List[int]
        wrong_inlinks = []  # type: List[Tuple[int,int]]
        wrong_outlinks = []  # type: List[Tuple[int,int]]

        reasons_incorrect_vertices = {}  # type: Dict[int, str]
        reasons_incorrect_inlinks = {}  # type: Dict[Tuple[int,int], str]
        reasons_incorrect_outlinks = {}  # type: Dict[Tuple[int,int], str]

        # Check that vertices have labels that are in the alphabet
        for v, class_name in list(vertices.items()):
            if class_name not in self.alphabet:
                wrong_vertices.append(v)
                reasons_incorrect_vertices[v] = 'Symbol {0} not in alphabet: class {1}.' \
                                                ''.format(v, class_name)

        # Check that all edges are allowed
        for from_id, to_id in edges:
            from_class_name, to_class_name = str(vertices[from_id]), str(vertices[to_id])
            if (from_class_name, to_class_name) not in self.rules:
                logging.debug('Wrong edge: {0} --> {1}, rules:\n{2}'
                              ''.format(from_class_name, to_class_name, pprint.pformat(self.rules)))

                wrong_inlinks.append((from_id, to_id))
                reasons_incorrect_inlinks[(from_id, to_id)] = 'Outlink {0} ({1}) -> {2} ({3}) not in alphabet.'.format(
                    from_class_name, from_id, to_class_name, to_id)

                wrong_outlinks.append((from_id, to_id))
                reasons_incorrect_outlinks[(from_id, to_id)] = 'Outlink {0} ({1}) -> {2} ({3}) not in alphabet.'.format(
                    from_class_name, from_id, to_class_name, to_id)
                if from_id not in wrong_vertices:
                    wrong_vertices.append(from_id)
                    reasons_incorrect_vertices[from_id] = 'Symbol {0} (class: {1}) participates ' \
                                                          'in wrong outlink: {2} ({3}) --> {4} ({5})' \
                                                          ''.format(from_id, vertices[from_id],
                                                                    from_class_name, from_id,
                                                                    to_class_name, to_id)
                if to_id not in wrong_vertices:
                    wrong_vertices.append(to_id)
                    reasons_incorrect_vertices[to_id] = 'Symbol {0} (class: {1}) participates ' \
                                                        'in wrong inlink: {2} ({3}) --> {4} ({5})' \
                                                        ''.format(to_id, vertices[to_id],
                                                                  from_class_name, from_id,
                                                                  to_class_name, to_id)

        # Check aggregate cardinality rules
        #  - build inlink and outlink dicts
        inlinks = {}
        outlinks = {}
        for v in vertices:
            outlinks[v] = set()
            inlinks[v] = set()
        for from_id, to_id in edges:
            outlinks[from_id].add(to_id)
            inlinks[to_id].add(from_id)

        # If there are not enough edges, the vertex itself is wrong
        # (and none of the existing edges are wrong).
        # Currently, if there are too many edges, the vertex itself
        # is wrong and none of the existing edges are marked.
        #
        # Future:
        # If there are too many edges, the vertex itself and *all*
        # the edges are marked as wrong (because any of them is the extra
        # edge, and it's easiest to just delete them and start parsing
        # again).
        logging.info('DependencyGrammar: checking outlink aggregate cardinalities'
                     '\n{0}'.format(pprint.pformat(outlinks)))
        for from_id in outlinks:
            from_class_name = vertices[from_id]
            if from_class_name not in self.outlink_aggregated_cardinalities:
                # Given vertex has no aggregate cardinality restrictions
                continue
            cmin, cmax = self.outlink_aggregated_cardinalities[from_class_name]
            logging.info('DependencyGrammar: checking outlink cardinality'
                         ' rule fulfilled for vertex {0} ({1}): should be'
                         ' within {2} -- {3}'.format(from_id, vertices[from_id], cmin, cmax))
            if not (cmin <= len(outlinks[from_id]) <= cmax):
                wrong_vertices.append(from_id)
                reasons_incorrect_vertices[from_id] = 'Symbol {0} (class: {1}) has {2} outlinks,' \
                                                      ' but grammar specifies {3} -- {4}.' \
                                                      ''.format(from_id, vertices[from_id],
                                                                len(outlinks[from_id]),
                                                                cmin, cmax)

        for to_id in inlinks:
            to_class_name = vertices[to_id]
            if to_class_name not in self.inlink_aggregated_cardinalities:
                continue
            cmin, cmax = self.inlink_aggregated_cardinalities[to_class_name]
            if not (cmin <= len(inlinks[to_id]) <= cmax):
                wrong_vertices.append(to_id)
                reasons_incorrect_vertices[to_id] = 'Symbol {0} (class: {1}) has {2} inlinks,' \
                                                    ' but grammar specifies {3} -- {4}.' \
                                                    ''.format(to_id, vertices[to_id],
                                                              len(inlinks[to_id]),
                                                              cmin, cmax)

        return wrong_vertices, wrong_inlinks, wrong_outlinks, reasons_incorrect_vertices, reasons_incorrect_inlinks, reasons_incorrect_outlinks

    def parse_dependency_grammar_rules(self, filename: str) -> \
            Tuple[List[Tuple[str, str]], Dict[str, Dict[str, Tuple[int, int]]], Dict[str, Dict[str, Tuple[int, int]]],
                  Dict[str, Tuple[int, int]], Dict[str, Tuple[int, int]]]:
        """Returns the rules stored in the given rule file.

        A dependency grammar rule file contains grammar lines,
        comment lines, and other lines. A grammar line is any line that
        contains the ``|`` symbol and does *not* have a ``#`` as the first
        non-whitespace symbol.

        Comment lines are those that have ``#`` as the first non-whitespace
        symbol. They are ignored, even if they contain ``|``. All other lines
        that do contain ``|`` are considered to be grammar lines.

        All lines that do not contain ``|`` are considered non-grammar lines
        and are ignored.

        :param filename: The path to the rule file.

        :returns: A quintuplet of the grammar rules.

            * ``rules``: a list of ``(from_class, to_class)`` tuples. Each rule tuple
              encodes that relationships leading from symbols of type ``from_class``
              to symbols of type ``to_class`` may exist.
            * ``inlink_cards``: a dictionary that encodes the range of permitted cardinalities
              for each RHS symbol of *inlinks* from the LHS symbols.
            * ``outlink_cards``: a dictionary that encodes the range of permitted cardinalities
              for each LHS of *outlinks* to the RHS symbols.
            * ``inlink_aggregate_cards``: A dict that holds for each RHS the range of
              permitted total inlink counts. E.g., a stem must always have at least one inlink.
            * ``outlink_aggregate_cards``: A dict that holds for each LHS the range
              of permitted total outlink counts. E.g., a full notehead must always have
              at least one outlink.

        """
        rules = []
        inlink_cardinalities = collections.OrderedDict()
        outlink_cardinalities = collections.OrderedDict()

        inlink_aggregated_cardinalities = collections.OrderedDict()
        outlink_aggregated_cardinalities = collections.OrderedDict()

        _invalid_lines = []
        with codecs.open(filename, 'r', 'utf-8') as hdl:
            for line_no, line in enumerate(hdl):
                l_rules, in_card, out_card, in_agg_card, out_agg_card = self.parse_dependency_grammar_line(
                    line)

                if not self.__validate_rules(l_rules):
                    _invalid_lines.append((line_no, line))

                rules.extend(l_rules)

                # Update cardinalities
                for lhs in out_card:
                    if lhs not in outlink_cardinalities:
                        outlink_cardinalities[lhs] = collections.OrderedDict()
                    outlink_cardinalities[lhs].update(out_card[lhs])

                for rhs in in_card:
                    if rhs not in inlink_cardinalities:
                        inlink_cardinalities[rhs] = collections.OrderedDict()
                    inlink_cardinalities[rhs].update(in_card[rhs])

                inlink_aggregated_cardinalities.update(in_agg_card)
                outlink_aggregated_cardinalities.update(out_agg_card)

        if len(_invalid_lines) > 0:
            logging.warning('DependencyGrammar.parse_rules(): Invalid lines'
                            ' {0}'.format(pprint.pformat(_invalid_lines)))

        return rules, inlink_cardinalities, outlink_cardinalities, inlink_aggregated_cardinalities, outlink_aggregated_cardinalities

    def parse_dependency_grammar_line(self, line: str) -> \
            Tuple[List[Tuple[str, str]], Dict[str, Dict[str, Tuple[int, int]]], Dict[
                str, Dict[str, Tuple[int, int]]], Dict[str, Tuple[int, int]], Dict[str, Tuple[int, int]]]:
        """Parse one dependency grammar line. See DependencyGrammar
        I/O documentation for the full format description of valid
        grammar lines.

        The grammar line specifies two kinds of information: which symbol
        classes may form relationships, and what the valid cardinalities
        for these relationships are. For instance, while ``time_signature``
        symbols have outlinks to ``numeral_X`` symbols, one numeral cannot
        be part of more than one time signature.

        A grammar line has a left-hand side (lhs) and a right-hand side (rhs),
        separated by the ``|`` symbol.

        (See :class:`DependencyGramamr` documentation for examples.)

        :param line: One line of a dependency grammar rule file.

        :returns: A quintuplet of:

            * ``rules``: a list of ``(from_class, to_class)`` tuples. Each rule tuple
              encodes that relationships leading from symbols of type ``from_class``
              to symbols of type ``to_class`` may exist.
            * ``inlink_cards``: a dictionary that encodes the range of permitted cardinalities
              for each RHS symbol of *inlinks* from the LHS symbols.
            * ``outlink_cards``: a dictionary that encodes the range of permitted cardinalities
              for each LHS of *outlinks* to the RHS symbols.
            * ``inlink_aggregate_cards``: A dict that holds for each RHS the range of
              permitted total inlink counts. E.g., a stem must always have at least one inlink.
            * ``outlink_aggregate_cards``: A dict that holds for each LHS the range
              of permitted total outlink counts. E.g., a full notehead must always have
              at least one outlink.

            For non-grammar lines (see :meth:`parse_dependency_grammar_rules`),
            this method returns empty data structures.

        """
        rules = []
        out_cards = collections.OrderedDict()
        in_cards = collections.OrderedDict()
        out_agg_cards = collections.OrderedDict()
        in_agg_cards = collections.OrderedDict()

        _no_rule_line_output = [], collections.OrderedDict(), collections.OrderedDict(), \
                               collections.OrderedDict(), collections.OrderedDict()
        if line.strip().startswith('#'):
            return _no_rule_line_output
        if len(line.strip()) == 0:
            return _no_rule_line_output
        if '|' not in line:
            return _no_rule_line_output

        # logging.info('DependencyGrammar: parsing rule line:\n\t\t{0}'
        #              ''.format(line.rstrip('\n')))
        lhs, rhs = line.strip().split('|', 1)
        lhs_tokens = lhs.strip().split()
        rhs_tokens = rhs.strip().split()

        # logging.info('DependencyGrammar: tokens lhs={0}, rhs={1}'
        #             ''.format(lhs_tokens, rhs_tokens))

        # Normal rule line? Aggregate cardinality line?
        _line_type = 'normal'
        if len(lhs) == 0:
            _line_type = 'aggregate_inlinks'
        if len(rhs) == 0:
            _line_type = 'aggregate_outlinks'

        logging.debug('Line {0}: type {1}, lhs={2}, rhs={3}'.format(line, _line_type, lhs, rhs))

        if _line_type == 'aggregate_inlinks':
            rhs_tokens = rhs.strip().split()
            for rt in rhs_tokens:
                token, rhs_cmin, rhs_cmax = self.parse_token(rt)
                for t in self.__matching_names(token):
                    in_agg_cards[t] = (rhs_cmin, rhs_cmax)
            logging.debug('DependencyGrammar: found inlinks: {0}'
                          ''.format(pprint.pformat(in_agg_cards)))
            return rules, out_cards, in_cards, in_agg_cards, out_agg_cards

        if _line_type == 'aggregate_outlinks':
            lhs_tokens = lhs.strip().split()
            for lt in lhs_tokens:
                token, lhs_cmin, lhs_cmax = self.parse_token(lt.strip())
                for t in self.__matching_names(token):
                    out_agg_cards[t] = (lhs_cmin, lhs_cmax)
            logging.debug('DependencyGrammar: found outlinks: {0}'
                          ''.format(pprint.pformat(out_agg_cards)))
            return rules, out_cards, in_cards, in_agg_cards, out_agg_cards

        # Normal line that defines a left-hand side and a right-hand side

        lhs_symbols = []
        # These cardinalities apply to all left-hand side tokens,
        # for edges leading to any of the right-hand side tokens.
        lhs_cards = collections.OrderedDict()
        for l in lhs_tokens:
            token, lhs_cmin, lhs_cmax = self.parse_token(l)
            all_tokens = self.__matching_names(token)
            lhs_symbols.extend(all_tokens)
            for t in all_tokens:
                lhs_cards[t] = (lhs_cmin, lhs_cmax)

        rhs_symbols = []
        rhs_cards = collections.OrderedDict()
        for r in rhs_tokens:
            token, rhs_cmin, rhs_cmax = self.parse_token(r)
            all_tokens = self.__matching_names(token)
            rhs_symbols.extend(all_tokens)
            for t in all_tokens:
                rhs_cards[t] = (rhs_cmin, rhs_cmax)

        # Build the outputs from the cartesian product
        # of left-hand and right-hand tokens.
        for l in lhs_symbols:
            if l not in out_cards:
                out_cards[l] = collections.OrderedDict()
            for r in rhs_symbols:
                if r not in in_cards:
                    in_cards[r] = collections.OrderedDict()

                rules.append((l, r))
                out_cards[l][r] = lhs_cards[l]
                in_cards[r][l] = rhs_cards[r]

        # Fixed rule ordering
        rules = sorted(rules)

        return rules, in_cards, out_cards, in_agg_cards, out_agg_cards

    def parse_token(self, token: str) -> Tuple[str, int, int]:
        """Parse one ``*.deprules`` file token. See class documentation for
        examples.

        :param token: One token of a ``*.deprules`` file.

        :return: token, cmin, cmax
        """
        token = str(token)
        cmin, cmax = 0, self._MAX_CARDINALITY
        if '{' not in token:
            token = token
        else:
            token, cardinality = token[:-1].split('{')
            if ',' not in cardinality:
                cmin, cmax = int(cardinality), int(cardinality)
            else:
                cmin_string, cmax_string = cardinality.split(',')
                if len(cmin_string) > 0:
                    cmin = int(cmin_string)
                if len(cmax_string) > 0:
                    cmax = int(cmax_string)
        return token, cmin, cmax

    def __matching_names(self, token: str) -> List[str]:
        """Returns the list of alphabet symbols that match the given
        name (regex, currently can process one '*' wildcard).

        :param token: The symbol name (pattern) to expand.

        :returns: A list of matching names. Empty list if no name matches.
        """
        if not self.__has_wildcard(token):
            return [token]

        wildcard_idx = token.index(self.WILDCARD)
        prefix = token[:wildcard_idx]
        if wildcard_idx < len(token) - 1:
            suffix = token[wildcard_idx + 1:]
        else:
            suffix = ''

        matching_names = list(self.alphabet)
        if len(prefix) > 0:
            matching_names = [n for n in matching_names if n.startswith(prefix)]
        if len(suffix) > 0:
            matching_names = [n for n in matching_names if n.endswith(suffix)]

        return matching_names

    def __validate_rules(self, rules: List[Tuple[str, str]]) -> bool:
        """Check that all the rules are valid under the current alphabet."""
        missing_heads = set()
        missing_children = set()
        for h, ch in rules:
            if h not in self.alphabet:
                missing_heads.add(h)
            if ch not in self.alphabet:
                missing_children.add(ch)

        if (len(missing_heads) + len(missing_children)) > 0:
            logging.warning('DependencyGrammar.validate_rules: missing heads '
                            '{0}, children {1}'
                            ''.format(missing_heads, missing_children))
            return False
        else:
            return True

    def __has_wildcard(self, name: str) -> bool:
        return self.WILDCARD in name



import copy
import itertools
import logging
from typing import List, Union, Tuple, Optional, Any

import numpy
from math import ceil

from mung.utils import compute_connected_components


class Node(object):
    """One annotated object.

    The Node represents one instance of an annotation. It implements
    the following attributes:

    * ``node_id``: the unique number of the given annotation instance in the set
      of annotations encoded in the containing `NodeList`.
    * ``dataset``: the name of the dataset this Node belongs to, e.g., MUSCIMA++_2.0
    * ``document``: the name of the document this Node belongs to, e.g., CVC-MUSCIMA_W-05_N-19_D-ideal
    * ``class_name``: the name of the label that was given to the annotation
      (this is the human-readable string such as ``notehead-full``).
    * ``top``: the vertical dimension (row) of the upper left corner pixel.
    * ``left``: the horizontal dimension (column) of the upper left corner pixel.
    * ``bottom``: the vertical dimension (row) of the lower right corner pixel + 1,
      so that you can index the corresponding image rows using
      ``img[c.top:c.bottom]``.
    * ``right``: the horizontal dimension (row) of the lower right corner pixel + 1,
      so that you can index the corresponding image columns using
      ``img[:, c.left:c.right]``.
    * ``width``: the amount of rows that the Node spans.
    * ``height``: the amount of columns that the Node spans.
    * ``mask``: a binary (0/1) numpy array that denotes the area within the
      Node's bounding box (specified by ``top``, ``left``, ``height``
      and ``width``) that the Node actually occupies. If the mask is
      ``None``, the object is understood to occupy the entire bounding box.
    * ``data``: a dictionary that can be empty, or can contain anything. It is
      generated from the optional ``<Data>`` element of a Node.

    Constructing a simple Node that consists of the "b"-like flat music
    notation symbol (never mind the ``unique_id`` for now):

    >>> top = 10
    >>> left = 15
    >>> height = 10
    >>> width = 4
    >>> mask = numpy.array([[1, 1, 0, 0],
    ...                     [1, 0, 0, 0],
    ...                     [1, 0, 0, 0],
    ...                     [1, 0, 0, 0],
    ...                     [1, 0, 1, 1],
    ...                     [1, 1, 1, 1],
    ...                     [1, 0, 0, 1],
    ...                     [1, 0, 1, 1],
    ...                     [1, 1, 1, 0],
    ...                     [0, 1, 0, 0]])
    >>> class_name = 'flat'
    >>> dataset = 'MUSCIMA-pp_2.0'
    >>> document = 'CVC-MUSCIMA_W-35_N-08_D-ideal'
    >>> node = Node(611, class_name=class_name,
    ...                top=top, left=left, height=height, width=width,
    ...                inlinks=[], outlinks=[],
    ...                mask=mask,
    ...                dataset=dataset, document=document)

    Nodes can also form graphs, using the following attributes:

    * ``outlinks``: Outgoing edges. A list of integers; it is assumed they are
      valid ``node_id`` within the same global/doc namespace.
    * ``inlinks``: Incoming edges. A list of integers; it is assumed they are
      valid ``node_id`` within the same global/doc namespace.

    So far, Node graphs do not support multiple relationship types.

    **Unique identification**

    The ``unique_id`` serves to identify the Node uniquely,
    at least within the MUSCIMA dataset system. (We anticipate further
    versions of the dataset, and need to plan for that.)

    To uniquely identify a Node, there are three "levels":

    * The "global", **dataset-level identification**: which dataset is this
      Node coming from? (For this dataset: ``MUSCIMA++_1.0``)
    * The "local", **document-level identification**: which document
      (within the given dataset) is this Node coming from?
      For MUSCIMA++ 1.0, this will usually be a string like
      ``CVC-MUSCIMA_W-35_N-08_D-ideal``, derived from the filename
      under which the Nodes containing the given Node
      is stored.
    * The **within-document identification**, which is the ``node_id``.

    These three components are joined together into one string by
    a delimiter: ``___``

    The full ``unique_id`` of a Node then might look like this::
    >>> node.unique_id
    'MUSCIMA-pp_2.0___CVC-MUSCIMA_W-35_N-08_D-ideal___611'

    And it consists of these three parts:

    >>> node.document
    'CVC-MUSCIMA_W-35_N-08_D-ideal'
    >>> node.dataset
    'MUSCIMA-pp_2.0'
    >>> node.id
    611

    **Nodes and images**

    Nodes and images are not tightly bound. This is because the same
    object can apply to multiple images: in the case of the CVC-MUSCIMA dataset,
    for example, the same Nodes are present both in the full image
    and in the staff-less image. The limitation here is that Nodes
    are based on exact pixels, so in order to retain validity, the images
    must correspond to each other exactly, as "layers".

    Because Nodes do not correspond to any given image, there is
    no facility in the data format to link them to a specific one. You have to
    take care of matching Node annotations to the right images by yourself.

    The ``Node`` class implements some interactions with images.

    To recover the area corresponding to a Node `c`, use:

    >>> image = numpy.array([]) #doctest: +SKIP
    >>> if node.mask is not None: crop = image[node.top:node.bottom, node.left:node.right] * node.mask  #doctest: +SKIP
    >>> if node.mask is None: crop = image[node.top:node.bottom, node.left:node.right]               #doctest: +SKIP

    Because this is clunky, we have implemented the following to get the crop:

    >>> crop = node.project_to(image)    #doctest: +SKIP

    And to get the Node projected onto the entire image:

    >>> crop = node.project_on(image)    #doctest: +SKIP

    Above, note the multiplicative role of the mask: while we typically would
    expect the mask to be binary, in principle, this is not strictly necessary.
    You could supply a different mask interpration, such as probabilistic.
    However, we strongly advise not to misuse this feature unless you have
    a really good reason; remember that the Node is supposed to represent
    an annotation of a given image. (One possible use for a non-binary mask
    that we can envision is aggregating multiple annotations of the same
    image.)

    For visualization, there is a more sophisticated method that renders
    the Node as a transparent colored transparent rectangle over
    an RGB image. (NOTE: this really changes the input image!)

    >>> import matplotlib.pyplot as plt #doctest: +SKIP
    >>> node.render(image)           #doctest: +SKIP
    >>> plt.imshow(image); plt.show() #doctest: +SKIP

    However, `Node.render()` currently does not support rendering
    the mask.

    **Disambiguating class names**

    Since the class names are present
    through the ``class_name`` attribute (``<ClassName>`` element),
    matching the list is no longer necessary for general understanding
    of the file. The NodeClasses file serves as a disambiguation tool:
    there may be multiple annotation projects that use the same names
    but maybe define them differently and use different guidelines,
    and their respective NodeClasses allow you to interpret the symbol
    names correctly, in light of the corresponding set of definitions.

    .. note::

        In MUSCIMarker, the NodeClasses is currently necessary to define
        how Nodes are displayed: their color. (All noteheads are red,
        all barlines are green, etc.) The other function, matching names
        to ``clsid``, has been superseeded by the ``class_name`` Node
        attribute.

    **Merging Nodes**

    To merge a list of Nodes into a new one, you need to:

    * Compute the new object's bounding box: ``compute_unifying_bounding_box()``
    * Compute the new object's mask: ``compute_unifying_mask()``
    * Determine the class_name and node_id of the new object.

    Since node_id and class_name of merges may depend on external settings
    and generally cannot be reliably determined from the merged
    objects themselves (e.g. the merge of a notehead and a stem
    should be a new note symbol), you need to supply them externally.
    However, the bounding box and mask can be determined. The bounding
    box is computed simply as the smallest bounding box that
    encompasses all the Nodes, and the mask is an OR operation
    over the individual masks (or None, if the Nodes don't
    have masks). Note that the merge cannot deal with a situation
    where only some of the objects have a mask.

    **Implementation notes on the mask**

    The mask is a numpy array that will be saved using run-length encoding.
    The numpy array is first flattened, then runs of successive 0's and 1's
    are encoded as e.g. ``0:10`` for a run of 10 zeros.

    How much space does this take?

    Objects tend to be relatively convex, so after flattening, we can expect
    more or less two runs per row (flattening is done in ``C`` order). Because
    each run takes (approximately) 5 characters, each mask takes roughly ``5 * n_rows``
    bytes to encode. This makes it efficient for objects wider than 5 pixels, with
    a compression ratio approximately ``n_cols / 5``.
    (Also, the numpy array needs to be made C-contiguous for that, which
    explains the ``order='C'`` hack in ``set_mask()``.)
    """

    # Delimits the Node UID fields (global, document namespaces, id)
    UID_DELIMITER = '___'
    DEFAULT_DATASET = 'MUSCIMA_DEFAULT_DATASET_PLACEHOLDER'
    DEFAULT_DOCUMENT = 'default-document'

    def __init__(self, id_: int,
                 class_name: str,
                 top: int,
                 left: int,
                 width: int,
                 height: int,
                 outlinks: List[int] = None,
                 inlinks: List[int] = None,
                 mask: numpy.ndarray = None,
                 dataset: str = None,
                 document: str = None,
                 data=None):
        self.__id = id_
        self.__class_name = class_name
        self.__top = top
        self.__left = left
        self.__width = width
        self.__height = height

        # The mask presupposes integer bounds.
        # Applied relative to Node bounds, not the whole image.
        self.__to_integer_bounds()
        self.__mask = None
        self.set_mask(mask)

        if inlinks is None:
            inlinks = []
        self.inlinks = inlinks  # type: List[int]

        if outlinks is None:
            outlinks = []
        self.outlinks = outlinks  # type: List[int]

        if dataset is None:
            dataset = self.DEFAULT_DATASET
        self.__dataset = dataset

        if document is None:
            document = self.DEFAULT_DOCUMENT
        self.__document = document

        self.is_selected = False

        if data is None:
            data = dict()
        self.data = data

    @property
    def unique_id(self) -> str:
        """Returns the ``unique_id`` of this Node

        >>> node = Node(0, "", 0, 0, 0, 0)
        >>> node.unique_id
        'MUSCIMA_DEFAULT_DATASET_PLACEHOLDER___default-document___0'
        """
        return self.UID_DELIMITER.join([self.dataset,
                                        self.document,
                                        str(self.id)])

    @staticmethod
    def parse_unique_id(uid: str) -> (str, str, int):
        """Parse a unique identifier. This breaks down the UID into the dataset name,
        document name, and id

        The delimiter is expected to be ``___``
        (kept as ``Node.UID_DELIMITER``)

        >>> Node.parse_unique_id('MUSCIMA++_2.0___CVC-MUSCIMA_W-05_N-19_D-ideal___424')
        ('MUSCIMA++_2.0', 'CVC-MUSCIMA_W-05_N-19_D-ideal', 424)

        :returns: ``global_namespace, document_namespace, id`` triplet.
            The namespaces are strings, ``id`` is an integer. If ``unique_id``
            is ``None``, returns ``None`` as ``id`` and expects it
            to be filled in from the caller Node instance.
        """
        if uid is None:
            global_namespace = Node.DEFAULT_DATASET
            document_namespace = Node.DEFAULT_DOCUMENT
            node_id = None
        else:
            global_namespace, document_namespace, node_id_string = uid.split(Node.UID_DELIMITER)
            node_id = int(node_id_string)
        return global_namespace, document_namespace, node_id

    @property
    def id(self) -> int:
        return self.__id

    def set_id(self, id_):
        self.__id = id_

    @property
    def class_name(self) -> str:
        return self.__class_name

    @property
    def dataset(self) -> str:
        return self.__dataset

    @property
    def document(self) -> str:
        return self.__document

    @property
    def top(self) -> int:
        """Row coordinate of upper left corner."""
        return self.__top

    @property
    def bottom(self) -> int:
        """Row coordinate 1 beyond bottom right corner, so that indexing
        in the form ``img[node.top:node.bottom]`` is possible."""
        return self.__top + self.__height

    @property
    def left(self) -> int:
        """Column coordinate of upper left corner."""
        return self.__left

    @property
    def right(self) -> int:
        """Column coordinate 1 beyond bottom right corner, so that indexing
        in the form ``img[:, node.left:node.right]`` is possible."""
        return self.__left + self.__width

    @property
    def width(self) -> int:
        return self.__width

    @property
    def height(self) -> int:
        return self.__height

    @property
    def bounding_box(self) -> Tuple[int, int, int, int]:
        """The ``top, left, bottom, right`` tuple of the Node's coordinates."""
        return self.top, self.left, self.bottom, self.right

    @property
    def middle(self) -> Tuple[int, int]:
        """Returns the integer representation of where the middle
        of the Node lies, as a ``(m_vert, m_horz)`` tuple.

        The integers just get rounded down.

        >>> node = Node(0,'', 10, 20, 30, 40)
        >>> node.middle
        (30, 35)
        """
        vertical_center = self.top + self.height // 2
        horizontal_center = self.left + self.width // 2
        return int(vertical_center), int(horizontal_center)

    @property
    def mask(self) -> numpy.ndarray:
        return self.__mask

    def set_mask(self, mask: numpy.ndarray):
        """Sets the Node's mask to the given array. Performs
        some compatibility checks: size, dtype (converts to ``uint8``)."""
        if mask is None:
            self.__mask = None
        else:
            # Check dimension
            t, l, b, r = self.round_bounding_box_to_integer(self.top,
                                                            self.left,
                                                            self.bottom,
                                                            self.right)
            if mask.shape != (b - t, r - l):
                raise ValueError('Mask shape {0} does not correspond'
                                 ' to integer shape {1} of Node.'
                                 ''.format(mask.shape, (b - t, r - l)))
            if str(mask.dtype) != 'uint8':
                logging.debug('Node.set_mask(): Supplied non-integer mask'
                              ' with dtype={0}'.format(mask.dtype))

            self.__mask = mask.astype('uint8')

    @staticmethod
    def round_bounding_box_to_integer(top: float, left: float, bottom: float, right: float) \
            -> (int, int, int, int):
        """Rounds off the Node bounds to the nearest integer
        so that no area is lost (e.g. bottom and right bounds are
        rounded up, top and left bounds are rounded down).

        Returns the rounded-off integers (top, left, bottom, right)
        as integers.

        >>> Node.round_bounding_box_to_integer(44.2, 18.9, 55.1, 92.99)
        (44, 18, 56, 93)
        >>> Node.round_bounding_box_to_integer(44, 18, 56, 92.99)
        (44, 18, 56, 93)

        """
        return int(top), int(left), int(ceil(bottom)), int(ceil(right))

    def project_to(self, image: numpy.ndarray):
        """This function returns the *crop* of the input image
        corresponding to the Node (incl. masking).
        Assumes zeros are background."""
        # Make a copy! We don't want to modify the original image by the mask.
        # Copy forced by the "* 1" part.
        crop = image[self.top:self.bottom, self.left:self.right] * 1
        if self.__mask is not None:
            crop *= self.__mask
        return crop

    def project_on(self, image: numpy.ndarray):
        """This function returns only those parts of the input image
        that correspond to the Node and masks out everything else
        with zeros. The dimension of the returned array is the same
        as of the input image. This function basically reconstructs
        the symbol as an indicator function over the pixels of
        the annotated image."""
        output = numpy.zeros(image.shape, image.dtype)
        crop = self.project_to(image)
        output[self.top:self.bottom, self.left:self.right] = crop
        return output

    def render(self, image: numpy.ndarray, alpha: float = 0.3,
               rgb: Tuple[float, float, float] = (1.0, 0.0, 0.0)) -> numpy.ndarray:
        """Renders itself upon the given image as a rectangle
        of the given color and transparency. Might help visualization.
        """
        color = numpy.array(rgb)
        logging.debug('Rendering object {0}, class_name {1}, t/b/l/r: {2}'
                      ''.format(self.id, self.class_name,
                                (self.top, self.bottom, self.left, self.right)))
        # logging.debug('Shape: {0}'.format((self.height, self.width, 3)))
        mask = numpy.ones((self.__height, self.__width, 3)) * color
        crop = image[self.top:self.bottom, self.left:self.right]
        # logging.debug('Mask done, creating crop')
        logging.debug('Shape: {0}. Got crop. Crop shape: {1}, img shape: {2}'
                      ''.format((self.__height, self.__width, 3), crop.shape, image.shape))
        mix = (crop + alpha * mask) / (1 + alpha)

        image[self.top:self.bottom, self.left:self.right] = mix
        return image

    def overlaps(self, bounding_box_or_node):
        # type: (Union[Tuple[int, int, int, int], Node]) -> bool
        """Check whether this Node overlaps the given bounding box or Node.

        >>> node = Node(0, 'test', 10, 100, height=20, width=10)
        >>> node.bounding_box
        (10, 100, 30, 110)
        >>> node.overlaps((10, 100, 30, 110))  # Exact match
        True
        >>> node.overlaps((0, 100, 8, 110))    # Row mismatch
        False
        >>> node.overlaps((10, 0, 30, 89))     # Column mismatch
        False
        >>> node.overlaps((0, 0, 8, 89))       # Total mismatch
        False
        >>> node.overlaps((9, 99, 31, 111))    # Encompasses Node
        True
        >>> node.overlaps((11, 101, 29, 109))  # Within Node
        True
        >>> node.overlaps((9, 101, 31, 109))   # Encompass horz., within vert.
        True
        >>> node.overlaps((11, 99, 29, 111))   # Encompasses vert., within horz.
        True
        >>> node.overlaps((11, 101, 31, 111))  # Corner within: top left
        True
        >>> node.overlaps((11, 99, 31, 109))   # Corner within: top right
        True
        >>> node.overlaps((9, 101, 29, 111))   # Corner within: bottom left
        True
        >>> node.overlaps((9, 99, 29, 109))    # Corner within: bottom right
        True

        """
        if isinstance(bounding_box_or_node, Node):
            t, l, b, r = bounding_box_or_node.bounding_box
        else:
            t, l, b, r = bounding_box_or_node
        # Does it overlap vertically? Includes situations where the Node is inside the bounding box.
        # Note that the bottom is +1 (fencepost), so the checks bottom vs. top need to be "less than",
        # not leq. If one object's top would be equal to the other's bottom, they would be touching,
        # not overlapping.
        if max(t, self.top) < min(b, self.bottom):
            if max(l, self.left) < min(r, self.right):
                return True
        return False

    def contains(self, bounding_box_or_node):
        # type: (Union[Tuple[int, int, int, int], Node]) -> bool
        """Check if this Node entirely contains the other bounding
        box (or, the other node's bounding box)."""
        if isinstance(bounding_box_or_node, Node):
            top, left, bottom, right = bounding_box_or_node.bounding_box
        else:
            top, left, bottom, right = bounding_box_or_node

        if self.top <= top <= bottom <= self.bottom:
            if self.left <= left <= right <= self.right:
                return True
        return False

    def bounding_box_intersection(self, bounding_box: Tuple[int, int, int, int]) \
            -> Optional[Tuple[int, int, int, int]]:
        """Returns the sub-bounding box of this Node intersecting with the given bounding box.
        If the intersection is empty, returns None.

        >>> node = Node(0, 'test', 10, 100, height=20, width=10)
        >>> node.bounding_box
        (10, 100, 30, 110)
        >>> other_bbox = 20, 100, 40, 105
        >>> node.bounding_box_intersection(other_bbox)
        (10, 0, 20, 5)
        >>> containing_bbox = 4, 55, 44, 115
        >>> node.bounding_box_intersection(containing_bbox)
        (0, 0, 20, 10)
        >>> contained_bbox = 12, 102, 22, 108
        >>> node.bounding_box_intersection(contained_bbox)
        (2, 2, 12, 8)
        >>> non_overlapping_bbox = 0, 0, 3, 3
        >>> node.bounding_box_intersection(non_overlapping_bbox) is None
        True

        """
        t, l, b, r = bounding_box

        out_top = max(t, self.top)
        out_bottom = min(b, self.bottom)
        out_left = max(l, self.left)
        out_right = min(r, self.right)

        if (out_top < out_bottom) and (out_left < out_right):
            return out_top - self.top, \
                   out_left - self.left, \
                   out_bottom - self.top, \
                   out_right - self.left
        else:
            return None

    def crop_to_mask(self):
        """Crops itself to the minimum bounding box that contains all
        its pixels, as determined by its mask.

        If the mask is all zeros, does not do anything, because
        at this point, the is_empty check should be invoked anyway
        in any situation where you care whether the object is empty
        or not (e.g. delete it after trimming).

        >>> mask = numpy.zeros((20, 10))
        >>> mask[5:15, 3:8] = 1
        >>> node = Node(0, 'test', 10, 100, width=10, height=20, mask=mask)
        >>> node.bounding_box
        (10, 100, 30, 110)
        >>> node.crop_to_mask()
        >>> node.bounding_box
        (15, 103, 25, 108)
        >>> node.height, node.width
        (10, 5)

        Assumes integer bounds, which is ensured during Node initialization.
        """
        if self.__mask is None:
            return

        mask_is_empty = self.__mask.sum() == 0
        if mask_is_empty:
            return

        # We know the object is not empty.

        # How many rows/columns to trim from top, bottom, etc.
        trim_top = -1
        for i in range(self.__mask.shape[0]):
            if self.__mask[i, :].sum() != 0:
                trim_top = i
                break

        trim_left = -1
        for j in range(self.__mask.shape[1]):
            if self.__mask[:, j].sum() != 0:
                trim_left = j
                break

        trim_bottom = -1
        for k in range(self.__mask.shape[0]):
            if self.__mask[-(k + 1), :].sum() != 0:
                trim_bottom = k
                break

        trim_right = -1
        for l in range(self.__mask.shape[1]):
            if self.__mask[:, -(l + 1)].sum() != 0:
                trim_right = l
                break

        logging.debug('Node.crop: Trimming top={0}, left={1},'
                      'bottom={2}, right={3}'
                      ''.format(trim_top, trim_left, trim_bottom, trim_right))

        # new bounding box relative to the current bounding box -- used to trim
        # the mask
        rel_t = trim_top
        rel_l = trim_left
        rel_b = self.__height - trim_bottom
        rel_r = self.__width - trim_right

        new_mask = self.__mask[rel_t:rel_b, rel_l:rel_r] * 1

        logging.debug('Node.crop: Old mask shape {0}, new mask shape {1}'
                      ''.format(self.__mask.shape, new_mask.shape))

        # new bounding box, relative to image -- used to compute the Node's position and size
        abs_t = self.top + trim_top
        abs_l = self.left + trim_left
        abs_b = self.bottom - trim_bottom
        abs_r = self.right - trim_right

        self.__top = abs_t
        self.__left = abs_l
        self.__height = abs_b - abs_t
        self.__width = abs_r - abs_l

        self.set_mask(new_mask)

    def __str__(self):
        """Format the Node as string representation. See the documentation
        of :module:`mung.io` for details."""
        lines = []
        lines.append('<Node>')
        lines.append('\t<Id>{0}</Id>'.format(self.id))
        lines.append('\t<ClassName>{0}</ClassName>'.format(self.class_name))
        lines.append('\t<Top>{0}</Top>'.format(self.top))
        lines.append('\t<Left>{0}</Left>'.format(self.left))
        lines.append('\t<Width>{0}</Width>'.format(self.__width))
        lines.append('\t<Height>{0}</Height>'.format(self.__height))

        mask_string = self.encode_mask()
        lines.append('\t<Mask>{0}</Mask>'.format(mask_string))

        if len(self.inlinks) > 0:
            inlinks_string = ' '.join(list(map(str, self.inlinks)))
            lines.append('\t<Inlinks>{0}</Inlinks>'.format(inlinks_string))
        if len(self.outlinks) > 0:
            outlinks_string = ' '.join(list(map(str, self.outlinks)))
            lines.append('\t<Outlinks>{0}</Outlinks>'.format(outlinks_string))

        data_string = self.encode_data()
        if data_string is not None:
            lines.append('\t<Data>\n{0}\n\t</Data>'.format(data_string))

        lines.append('</Node>')
        return '\n'.join(lines)

    def encode_mask(self, mode: str = 'rle') -> str:
        """Encode a binary array ``mask`` as a string, compliant
        with the Node format specification in :mod:`mung.io`.
        """
        if mode == 'rle':
            return self.encode_mask_rle(self.mask)
        elif mode == 'bitmap':
            return self.encode_mask_bitmap(self.mask)

    def encode_data(self) -> Optional[str]:
        if self.data is None:
            return None
        if len(self.data) == 0:
            return None

        lines = []
        for k, v in list(self.data.items()):
            vtype = 'str'
            vval = v
            if isinstance(v, int):
                vtype = 'int'
                vval = str(v)
            elif isinstance(v, float):
                vtype = 'float'
                vval = str(v)
            elif isinstance(v, list):
                vtype = 'list[str]'
                if len(v) > 0:
                    if isinstance(v[0], int):
                        vtype = 'list[int]'
                    elif isinstance(v[0], float):
                        vtype = 'list[float]'
                vval = ' '.join([str(vv) for vv in v])

            line = '\t\t<DataItem key="{0}" type="{1}">{2}</DataItem>' \
                   ''.format(k, vtype, vval)
            lines.append(line)

        return '\n'.join(lines)

    def data_display_text(self) -> str:
        if self.data is None:
            return '[No data]'
        if len(self.data) == 0:
            return '[No data]'

        lines = []
        for k, v in list(self.data.items()):
            lines.append('{0}:      {1}'.format(k, v))
        return '\n'.join(lines)

    @staticmethod
    def encode_mask_bitmap(mask: numpy.ndarray) -> str:
        """Encodes the mask array in a compact form. Returns 'None' if mask
        is None. If the mask is not None, uses the following algorithm:

        * Flatten the mask (then use width and height of Node for reshaping).
        * Record as string, with whitespace separator
        * Return resulting string
        """
        if mask is None:
            return 'None'
        # By default works in row-major order.
        # So we can just prescribe 'C' without losing data.
        mask_flat = mask.flatten(order='C')
        output = ' '.join(list(map(str, mask_flat)))
        return output

    @staticmethod
    def encode_mask_rle(mask: numpy.ndarray) -> str:
        """Encodes the mask array in Run-Length Encoding. Instead of
        having the bitmap ``0 0 1 1 1 0 0 0 1 1``, the RLE encodes
        the mask as ``0:2 1:3 0:3 1:2``. This is much more compact.

        Currently, the rows of the mask are not treated in any special
        way. The mask just gets flattened and then encoded.

        """
        if mask is None:
            return 'None'
        mask_flat = mask.flatten(order='C')

        output_strings = []
        current_run_type = 0
        current_run_length = 0
        for i in mask_flat:
            if i == current_run_type:
                current_run_length += 1
            else:
                s = '{0}:{1}'.format(current_run_type, current_run_length)
                output_strings.append(s)
                current_run_type = i
                current_run_length = 1
        s = '{0}:{1}'.format(current_run_type, current_run_length)
        output_strings.append(s)
        output = ' '.join(output_strings)
        return output

    @staticmethod
    def decode_mask(mask_string: str, shape) -> Optional[numpy.ndarray]:
        """Decodes a Node mask string into a binary
        numpy array of the given shape."""
        mode = Node.__determine_mask_mode(mask_string)
        if mode == 'rle':
            return Node.decode_mask_rle(mask_string, shape=shape)
        elif mode == 'bitmap':
            return Node.decode_mask_bitmap(mask_string, shape=shape)

    @staticmethod
    def __determine_mask_mode(mask_string: str):
        """If the mask string starts with '0:' or '1:', or generally
        if it contains a non-0 or 1 symbol, assume it is RLE."""
        mode = 'bitmap'
        if len(mask_string) < 3:
            mode = 'bitmap'
        elif ':' in mask_string[:3]:
            mode = 'rle'
        return mode

    @staticmethod
    def decode_mask_bitmap(mask_string: str, shape) -> Optional[numpy.ndarray]:
        """Decodes the mask array from the encoded form to the 2D numpy array."""
        if mask_string == 'None':
            return None
        try:
            values = list(map(float, mask_string.split()))
        except ValueError:
            logging.info(
                'Node.decode_mask_bitmap() Cannot decode mask values:\n{0}'.format(mask_string))
            raise
        mask = numpy.array(values).reshape(shape)
        return mask

    @staticmethod
    def decode_mask_rle(mask_string: str, shape) -> Optional[numpy.ndarray]:
        """Decodes the mask array from the RLE-encoded form
        to the 2D numpy array.
        """
        if mask_string == 'None':
            return None

        values = []
        for kv in mask_string.split(' '):
            k_string, v_string = kv.split(':')
            k, v = int(k_string), int(v_string)
            vs = [k for _ in range(v)]
            values.extend(vs)

        mask = numpy.array(values).reshape(shape)
        return mask

    def join(self, other):
        """Node "addition": performs an OR on this
        and the ``other`` Nodes' masks and bounding boxes,
        and assigns to this Node the result. Merges
        also the inlinks and outlinks.

        Works only if the document spaces for both Nodes
        are the same. (Otherwise changes nothing.)

        The ``class_name`` of the ``other`` is ignored.
        """
        if self.document != other.document:
            logging.warning(
                "Trying to join Node from different documents, which is forbidden. Skipping join.")
            return

        # Get combined bounding box
        new_top = min(self.top, other.top)
        new_left = min(self.left, other.left)
        new_bottom = max(self.bottom, other.bottom)
        new_right = max(self.right, other.right)

        new_height = new_bottom - new_top
        new_width = new_right - new_left

        # Create mask of corresponding size
        new_mask = numpy.zeros((new_height, new_width), dtype=self.__mask.dtype)

        # Find coordinates where to paste the masks
        spt = self.top - new_top
        spl = self.left - new_left
        opt = other.top - new_top
        opl = other.left - new_left

        # Paste the masks into these places
        new_mask[spt:spt + self.__height, spl:spl + self.__width] += self.__mask
        new_mask[opt:opt + other.height, opl:opl + other.width] += other.mask

        # Normalize mask value
        new_mask[new_mask != 0] = 1

        # Assign the new variables to this Node
        self.__top = new_top
        self.__left = new_left
        self.__height = new_height
        self.__width = new_width
        self.__mask = new_mask

        # Add inlinks and outlinks (check for multiple and self-reference)
        for o in other.outlinks:
            if (o not in self.outlinks) and (o != self.id):
                self.outlinks.append(o)
        for i in other.inlinks:
            if (i not in self.inlinks) and (i != self.id):
                self.inlinks.append(i)

    def get_outlink_objects(self, nodes):
        # type: (List[Node]) -> List[Node]
        """Out of the given ``nodes`` list, return a list
        of those to which this Node has outlinks.
        Can deal with Nodes from multiple documents.
        """
        return self.__check_nodes_that_have_links(self.outlinks, nodes)

    def get_inlink_objects(self, nodes):
        # type: (List[Node]) -> List[Node]
        """Out of the given ``nodes`` list, return a list
        of those from which this node has inlinks
        Can deal with Nodes from multiple documents.
        """
        return self.__check_nodes_that_have_links(self.inlinks, nodes)

    def __check_nodes_that_have_links(self, links, nodes):
        # type: (List[int], List[Node]) -> List[Node]
        output = []
        if len(links) == 0:
            return output

        link_set = frozenset(links)

        for node in nodes:
            if node.document != self.document:
                continue
            if node.id in link_set:
                output.append(node)
            if len(output) == len(self.inlinks):
                break
        return output

    def translate(self, down: int = 0, right: int = 0):
        """Move the Node down and right by the given amount of pixels."""
        self.__top += down
        self.__left += right

    def scale(self, zoom: float = 1.0):
        """Re-compute the Node with the given scaling factor."""
        mask = self.__mask * 1.0
        import skimage.transform
        new_mask_shape = max(int(self.__height * zoom), 1), max(int(self.__width * zoom), 1)
        new_mask = skimage.transform.resize(mask,
                                            output_shape=new_mask_shape)
        new_mask[new_mask >= 0.5] = 1
        new_mask[new_mask < 0.5] = 0
        new_mask = new_mask.astype('uint8')

        new_height, new_width = new_mask.shape
        new_top = int(self.top * zoom)
        new_left = int(self.left * zoom)

        self.__top = new_top
        self.__left = new_left
        self.__height = new_height
        self.__width = new_width
        self.__mask = new_mask

    def __to_integer_bounds(self):
        """Ensures that the Node has an integer position and size.
        (This is important whenever you want to use a mask, and reasonable
        whenever you do not need sub-pixel resolution...)
        """
        bounding_box = self.bounding_box
        top, left, bottom, right = self.round_bounding_box_to_integer(*bounding_box)
        height = bottom - top
        width = right - left

        self.__top = top
        self.__left = left
        self.__height = height
        self.__width = width

    def distance_to(self, node) -> Any:
        """Computes the distance between this node and another node.
        Their minimum vertical and horizontal distances are each taken
        separately, and the euclidean norm is computed from them."""
        if self.document != node.document:
            logging.warning('Cannot compute distances between Nodes'
                            ' from different documents! ({0} vs. {1})'
                            ''.format(self.document, node.document))

        if (self.top <= node.top <= self.bottom) or (node.top <= self.top <= node.bottom):
            delta_vert = 0
        elif self.top < node.top:
            delta_vert = node.top - self.bottom
        else:
            delta_vert = self.top - node.bottom

        if (self.left <= node.left <= self.right) or (node.left <= self.left <= node.right):
            delta_horz = 0
        elif self.left < node.left:
            delta_horz = node.left - self.right
        else:
            delta_horz = self.left - node.right

        return numpy.sqrt(delta_vert ** 2 + delta_horz ** 2)

    def compute_recall_precision_fscore_on_mask(self, other_node):
        # type: (Node) -> Tuple[float, float, float]
        """Compute the recall, precision and f-score of the predicted
        Node's mask against another node's mask."""

        if bounding_box_intersection(self.bounding_box, other_node.bounding_box) is None:
            return 0.0, 0.0, 0.0

        mask_intersection = compute_unifying_mask([(self), (other_node)], intersection=False)

        gt_pasted_mask = mask_intersection * 1
        t, l, b, r = compute_unifying_bounding_box([self, other_node])
        h, w = b - t, r - l
        ct, cl, cb, cr = self.top - t, \
                         self.left - l, \
                         h - (b - self.bottom), \
                         w - (r - self.right)
        gt_pasted_mask[ct:cb, cl:cr] += self.mask
        gt_pasted_mask[gt_pasted_mask != 0] = 1

        pred_pasted_mask = mask_intersection * 1
        t, l, b, r = other_node.bounding_box
        h, w = b - t, r - l
        ct, cl, cb, cr = other_node.top - t, \
                         other_node.left - l, \
                         h - (b - other_node.bottom), \
                         w - (r - other_node.right)
        pred_pasted_mask[ct:cb, cl:cr] += other_node.mask
        pred_pasted_mask[pred_pasted_mask != 0] = 1

        true_positives = float(mask_intersection.sum())
        false_positives = pred_pasted_mask.sum() - true_positives
        false_negatives = gt_pasted_mask.sum() - true_positives
        recall = true_positives / (true_positives + false_negatives)
        precision = true_positives / (true_positives + false_positives)
        f_score = (2 * recall * precision) / (recall + precision)

        return recall, precision, f_score


##############################################################################


def split_node_by_its_connected_components(node: Node, next_node_id: int) -> List[Node]:
    """Split the Node into one object per connected component
    of the mask. All inlinks/outlinks are retained in all the newly
    created Nodes, and the old object is not changed.
    If there is only one connected component, the object is returned unchanged
    in a list with one entry.

    A ``id`` must be provided at which to start numbering the newly
    created Nodes.

    The ``data`` attribute is also retained.
    """
    # "Safety margin"
    canvas = numpy.zeros((node.mask.shape[0] + 2, node.mask.shape[1] + 2))
    canvas[1:-1, 1:-1] = node.mask
    number_of_connected_components, labels, bounding_boxes = compute_connected_components(canvas)

    logging.info('Node.split(): {0} connected components, bounding boxes: {1}'
                 .format(number_of_connected_components, bounding_boxes))

    if len(bounding_boxes) == 1:
        return [node]

    output = []

    for label, (top, left, bottom, right) in list(bounding_boxes.items()):
        # Background in compute_connected_components() doesn't work?
        if label == 0:
            continue

        height = bottom - top
        width = right - left
        m_label = (labels == label).astype('uint8')
        m = m_label[top:bottom, left:right]
        top = top + node.top - 1
        left = left + node.left - 1
        node_id = next_node_id
        inlinks = copy.deepcopy(node.inlinks)
        outlinks = copy.deepcopy(node.outlinks)
        data = copy.deepcopy(node.data)
        dataset = node.dataset
        document = node.document

        new_node = Node(node_id, node.class_name, top, left, width, height,
                        inlinks=inlinks, outlinks=outlinks,
                        mask=m, data=data, dataset=dataset, document=document)
        output.append(new_node)

        next_node_id += 1

    return output


def merge_nodes(first_node: Node, second_node: Node, class_name: str, id_: int) -> Node:
    """Merge the given Nodes with respect to the other.
    Returns a new Node (without modifying any of the inputs)."""
    return merge_multiple_nodes([first_node, second_node], class_name, id_)


def merge_multiple_nodes(nodes: List[Node], class_name: str, id_: int) -> Node:
    """Merge multiple nodes. Does not modify any of the inputs."""
    if len(set([c.document for c in nodes])) > 1:
        raise ValueError('Cannot merge Nodes from different documents!')
    merged_top, merged_left, merged_bottom, merged_right = compute_unifying_bounding_box(nodes)
    merged_height, merged_width = merged_bottom - merged_top, merged_right - merged_left
    merged_mask = compute_unifying_mask(nodes)
    merged_inlinks, merged_outlinks = merge_inlinks_and_outlinks_to_nodes_outside_of_this_list(
        nodes)

    dataset = nodes[0].dataset
    document = nodes[0].document

    output = Node(id_, class_name,
                  top=merged_top, left=merged_left, height=merged_height, width=merged_width,
                  mask=merged_mask,
                  inlinks=merged_inlinks, outlinks=merged_outlinks,
                  dataset=dataset, document=document)
    return output


def compute_unifying_bounding_box(nodes: List[Node]) -> (int, int, int, int):
    """ Computes the union bounding box of multiple nodes """
    top, left, bottom, right = numpy.inf, numpy.inf, -1, -1
    for node in nodes:
        top = min(top, node.top)
        left = min(left, node.left)
        bottom = max(bottom, node.bottom)
        right = max(right, node.right)

    it, il, ib, ir = int(top), int(left), int(bottom), int(right)
    if (it != top) or (il != left) or (ib != bottom) or (ir != right):
        logging.warning('Merged bounding box does not consist of integers!'
                        ' {0}'.format((top, left, bottom, right)))

    return it, il, ib, ir


def compute_unifying_mask(nodes: List[Node], intersection=False) -> Optional[numpy.ndarray]:
    """ Merges the masks of the given Nodes into one. Masks are combined by an OR operation.

    >>> c1 = Node(0, 'name', 10, 10, 4, 1, mask=numpy.ones((1, 4), dtype='uint8'))
    >>> c2 = Node(1, 'name', 11, 10, 6, 1, mask=numpy.ones((1, 6), dtype='uint8'))
    >>> c3 = Node(2, 'name', 9, 14,  2, 4, mask=numpy.ones((4, 2), dtype='uint8'))
    >>> nodes = [c1, c2, c3]
    >>> m1 = compute_unifying_mask(nodes)
    >>> m1.shape
    (4, 6)
    >>> print(m1)
    [[0 0 0 0 1 1]
     [1 1 1 1 1 1]
     [1 1 1 1 1 1]
     [0 0 0 0 1 1]]

    Mask behavior: if at least one of the Nodes has a mask, then
    masking behavior is activated. The masks are combined using OR: any
    pixel of the resulting merged Node that corresponds to a True
    mask pixel in one of the input Nodes will get a True mask value,
    all others (ie. including all intermediate areas) will get a False.

    If no input Node has a mask, then the resulting Node also will not have a mask.

    If some Nodes have masks and some don't, this call with throw an error.

    :param nodes: The list of nodes whose masks will be merged

    :param intersection: Instead of a union, return the mask
        intersection: only those pixels which are common to all
        the Nodes.
    """
    no_node_has_a_mask = len([c for c in nodes if c.mask is not None]) == 0
    if no_node_has_a_mask:
        return None

    for node in nodes:
        if node.mask is None:
            # Some nodes have masks and some don't
            raise ValueError('Cannot deal with a mix of masked and non-masked Nodes.')

    top, left, bottom, right = compute_unifying_bounding_box(nodes)
    height = bottom - top
    width = right - left
    output_mask = numpy.zeros((height, width), dtype=nodes[0].mask.dtype)
    for node in nodes:
        ct, cl, cb, cr = node.top - top, node.left - left, height - (
                bottom - node.bottom), width - (right - node.right)
        output_mask[ct:cb, cl:cr] += node.mask

    if intersection:
        output_mask[output_mask < len(nodes)] = 0
        output_mask[output_mask != 0] = 1
    else:
        output_mask[output_mask > 0] = 1
    return output_mask


def merge_inlinks_and_outlinks_to_nodes_outside_of_this_list(nodes: List[Node]) \
        -> Tuple[List[int], List[int]]:
    """Collect all inlinks and outlinks of the given set of Nodes
    to Nodes outside of this set. The rationale for this is that
    these given ``nodes`` will be merged into one, so relationships
    within the set would become loops and disappear.

    (Note that this is not sufficient to update the relationships upon
    a merge, because the affected Nodess *outside* the given set
    will need to have their inlinks/outlinks redirected to the new object.)

    :returns: A tuple of lists: ``(inlinks, outlinks)``
    """
    all_node_ids = frozenset([node.id for node in nodes])
    outlinks = []
    inlinks = []
    for c in nodes:
        # No duplicates
        outlinks.extend([o for o in c.outlinks
                         if (o not in all_node_ids) and (o not in outlinks)])
        inlinks.extend([i for i in c.inlinks
                        if (i not in all_node_ids) and (i not in inlinks)])
    return inlinks, outlinks


def merge_node_lists_from_multiple_documents(node_lists: List[List[Node]]) -> List[Node]:
    """Combines the Node lists from different documents
    into one list, so that inlink/outlink references still work.
    This is useful only if you want to merge two documents
    into one (e.g., if your annotators worked on different "layers"
    of data, and you want to merge these annotations).

    This just means shifting the ``id`` (and thus inlinks
    and outlinks). It is assumed the lists pertain to the same
    image. Uses deepcopy to avoid exposing the original lists
    to modification through the merged list.

    Currently cannot handle precedence edges.

    """
    max_node_ids = [max([node.id for node in c_list]) for c_list in node_lists]
    min_node_ids = [min([node.id for node in c_list]) for c_list in node_lists]
    shift_by = [0] + [sum(max_node_ids[:i]) - min_node_ids[i] + 1 for i in
                      range(1, len(max_node_ids))]

    new_lists = []
    for nodes, s in zip(node_lists, shift_by):
        new_list = []
        for node in nodes:
            new_node = copy.deepcopy(node)
            new_id = node.id + s
            new_node.set_id(new_id)

            # Graph handling
            new_node.inlinks = [i + s for i in node.inlinks]
            new_node.outlinks = [o + s for o in node.outlinks]

            new_list.append(new_node)
        new_lists.append(new_list)

    output = list(itertools.chain(*new_lists))

    return output


def link_nodes(from_node: Node, to_node: Node, check_that_nodes_have_the_same_document: bool = True):
    """Add a relationship from one node to the other. Updates the nodes in-place.

    If the objects are already linked, does nothing.
    """
    if from_node.document != to_node.document:
        if check_that_nodes_have_the_same_document:
            raise ValueError('Cannot link two Nodes that are')
        else:
            logging.warning('Attempting to link Nodes from two different'
                            ' docments. From: {0}, to: {1}'
                            ''.format(from_node.document, to_node.document))

    if (to_node.id not in from_node.outlinks) and (from_node.id in to_node.inlinks):
        logging.warning('Malformed object graph in document {0}:'
                        ' Relationship {1} --> {2} already exists as inlink,'
                        ' but not as outlink!.'
                        ''.format(from_node.document, from_node.id, to_node.id))
    from_node.outlinks.append(to_node.id)
    to_node.inlinks.append(from_node.id)


def bounding_box_intersection(first_bounding_box: Tuple[int, int, int, int],
                              second_bounding_box: Tuple[int, int, int, int]) -> Optional[
    Tuple[int, int, int, int]]:
    """Returns the t, l, b, r coordinates of the sub-bounding box
    of bbox_this that is also inside bbox_other.
    If the bounding boxes do not overlap, returns None."""
    t, l, b, r = second_bounding_box
    tt, tl, tb, tr = first_bounding_box

    out_top = max(t, tt)
    out_bottom = min(b, tb)
    out_left = max(l, tl)
    out_right = min(r, tr)

    if (out_top < out_bottom) and (out_left < out_right):
        return out_top - tt, \
               out_left - tl, \
               out_bottom - tt, \
               out_right - tl
    else:
        return None


def bounding_box_dice_coefficient(first_bounding_box: Tuple[int, int, int, int],
                                  second_bounding_box: Tuple[int, int, int, int],
                                  vertical: bool = False,
                                  horizontal: bool = False) -> float:
    """Compute the Dice coefficient (intersection over union)
    for the given two bounding boxes.

    :param vertical: If set, will only return vertical IoU.

    :param horizontal: If set, will only return horizontal IoU.
        If both vertical and horizontal are set, will return
        normal IoU, as if they were both false.
    """
    t_t, t_l, t_b, t_r = first_bounding_box
    o_t, o_l, o_b, o_r = second_bounding_box

    u_t, i_t = min(t_t, o_t), max(t_t, o_t)
    u_l, i_l = min(t_l, o_l), max(t_l, o_l)
    u_b, i_b = max(t_b, o_b), min(t_b, o_b)
    u_r, i_r = max(t_r, o_r), min(t_r, o_r)

    u_vertical = max(0, u_b - u_t)
    u_horizontal = max(0, u_r - u_l)

    i_vertical = max(0, i_b - i_t)
    i_horizontal = max(0, i_r - i_l)

    if vertical and not horizontal:
        if u_vertical == 0:
            return 0.0
        else:
            return i_vertical / u_vertical
    elif horizontal and not vertical:
        if u_horizontal == 0:
            return 0.0
        else:
            return i_horizontal / u_horizontal
    else:
        if (u_horizontal == 0) or (u_vertical == 0):
            return 0.0
        else:
            return (i_horizontal * i_vertical) / (u_horizontal * u_vertical)


def draw_nodes_on_empty_canvas(nodes: List[Node], margin: int = 10) -> Tuple[
    numpy.ndarray, Tuple[int, int]]:
    """Draws all the given Nodes onto a zero background.
    The size of the canvas adapts to the Nodes, with the    given margin.

    Also returns the top left corner coordinates w.r.t. Nodes' bounding boxes.
    """

    # margin is used to avoid the stafflines touching the edges,
    # which could perhaps break some assumptions down the line.
    top, left, bottom, right = compute_unifying_bounding_box(nodes)
    top_with_margin, left_with_margin, bottom_with_margin, right_with_margin = \
        max(0, top - margin), max(0, left - margin), bottom + margin, right + margin

    canvas = numpy.zeros(
        (bottom_with_margin - top_with_margin, right_with_margin - left_with_margin))

    for node in nodes:
        canvas[node.top - top_with_margin:node.bottom - top_with_margin,
        node.left - left_with_margin:node.right - left_with_margin] = node.mask * 1

    canvas[canvas != 0] = 1

    return canvas, (top_with_margin, left_with_margin)



#!/usr/bin/env python
"""``analyze_annotation.py`` is a script that analyzes annotation results.

For an overview of command-line options, call::

  analyze_annotation.py -h


Functionality
-------------

* Count symbols
* Count symbol classes
* Compute symbol parameters per class (size, morphological features..?) [NOT IMPLEMENTED]

* Count relationships
* Count relationship classes
* Compute relationship parameters per class pair

"""
import argparse
import collections
import logging
import pprint
import time

import operator

from typing import List, Tuple, Dict, Any

from mung.io import read_nodes_from_file, get_edges
from mung.node import merge_node_lists_from_multiple_documents, Node


def compute_node_statistics(nodes: List[Node], edges: List[Tuple[int, int]] = None) -> Dict[str, Any]:
    stats = collections.OrderedDict()

    # Count Nodes
    stats['number_of_nodes'] = len(nodes)

    # Count Nodes by class
    number_of_nodes_by_class = collections.defaultdict(int)
    for node in nodes:
        number_of_nodes_by_class[node.class_name] += 1
    stats['number_of_nodes_by_class'] = number_of_nodes_by_class
    stats['number_of_distinct_nodes'] = len(number_of_nodes_by_class)

    if edges is not None:
        # Count relationships
        id_to_node_mapping = {node.id: node for node in nodes}
        stats['number_of_relationships'] = len(edges)
        number_of_relationships_by_class = collections.defaultdict(int)
        for edge in edges:
            from_node_id, to_node_id = edge
            from_node = id_to_node_mapping[from_node_id].class_name
            to_node = id_to_node_mapping[to_node_id].class_name
            number_of_relationships_by_class[(from_node, to_node)] += 1
        stats['number_of_relationships_by_class'] = number_of_relationships_by_class
        stats['number_of_relationships_distinct'] = len(number_of_relationships_by_class)

    return stats


def print_statistics(stats: Dict[str, Any]):
    # For now, just pretty-print. That means reformatting the insides
    # of the stats.
    print_stats = list()

    if 'number_of_nodes' in stats:
        print_stats.append(('number_of_nodes', stats['number_of_nodes']))
    if 'number_of_nodes_by_class' in stats:
        print_stats.append(('number_of_nodes_by_class',
                            sorted(list(stats['number_of_nodes_by_class'].items()),
                                   key=operator.itemgetter(1),
                                   reverse=True)
                            ))
    if 'number_of_distinct_nodes' in stats:
        print_stats.append(('number_of_distinct_nodes',
                            stats['number_of_distinct_nodes']))

    if 'number_of_relationships' in stats:
        print_stats.append(('number_of_relationships', stats['number_of_relationships']))
    if 'number_of_relationships_by_class' in stats:
        print_stats.append(('number_of_relationships_by_class',
                            sorted(list(stats['number_of_relationships_by_class'].items()),
                                   key=operator.itemgetter(1),
                                   reverse=True)))
    if 'number_of_relationships_distinct' in stats:
        print_stats.append(('number_of_relationships_distinct',
                            stats['number_of_relationships_distinct']))

    pprint.pprint(print_stats)


##############################################################################

def build_argument_parser():
    parser = argparse.ArgumentParser(description=__doc__, add_help=True,
                                     formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('-i', '--input', action='store', nargs='+',
                        required=True,
                        help='List of input NodeList files.')
    parser.add_argument('-e', '--emit', action='store', default='print',
                        choices=['print'],
                        help='How should the analysis results be presented?')

    parser.add_argument('-v', '--verbose', action='store_true',
                        help='Turn on INFO messages.')
    parser.add_argument('--debug', action='store_true',
                        help='Turn on DEBUG messages.')

    return parser


if __name__ == '__main__':
    parser = build_argument_parser()
    args = parser.parse_args()

    if args.verbose:
        logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.INFO)
    if args.debug:
        logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.DEBUG)

    logging.info('Starting main...')
    _start_time = time.clock()

    # Parse individual Node lists.
    node_lists = []
    number_of_parsed_nodes = 0
    for i, f in enumerate(args.input):
        node_list = read_nodes_from_file(f)
        node_lists.append(node_list)

        # Logging progress
        number_of_parsed_nodes += len(node_list)
        if i % 10 == 0 and i > 0:
            _time_parsing = time.clock() - _start_time
            nodes_per_second = number_of_parsed_nodes / _time_parsing
            logging.info('Parsed {0} Nodes in {1:.2f} s ({2:.2f} objs/s)'
                         ''.format(number_of_parsed_nodes,
                                   _time_parsing, nodes_per_second))

    # Merge the Node lists into one.
    # This is done so that the resulting object graph can be manipulated
    # at once, without id clashes.
    merged_node_list = merge_node_lists_from_multiple_documents(node_lists)

    edges = get_edges(merged_node_list)

    _parse_end_time = time.clock()
    logging.info('Parsing took {0:.2f} s'.format(_parse_end_time - _start_time))

    ##########################################################################
    # Analysis

    # Here's where the results are stored, for export into various
    # formats. (Currently, we only print them.)
    statistics = compute_node_statistics(merged_node_list, edges=edges)

    ##########################################################################
    # Export
    if args.emit == 'print':
        print_statistics(statistics)

    _end_time = time.clock()
    logging.info('analyze_annotations.py done in {0:.3f} s'
                 ''.format(_end_time - _start_time))




"""This module acts as an abstraction over the CVC-MUSCIMA dataset.

It mostly implements utility functions, like getting the absolute
path to a specific image in the CVC-MUSCIMA dataset, specified
by the writer, number, distortion, and mode.

Environmental variables
-----------------------

* ``CVC_MUSCIMA_ROOT``
* ``MUSCIMA_PLUSPLUS_ROOT``

The dataset root environmental variables are used as default roots
for retrieving the dataset files. If they are not set, you will
have to supply the roots to the respective functions that manipulate
these layers of MUSCIMA++.
"""
import logging
import os
from typing import Optional


def _get_cvc_muscima_root() -> Optional[str]:
    if 'CVC_MUSCIMA_ROOT' in os.environ:
        cvc_muscima_root = os.environ['CVC_MUSCIMA_ROOT']
        return cvc_muscima_root
    else:
        logging.info('muscima.dataset: environmental variable CVC_MUSCIMA_ROOT not defined.')
        return None


CVC_MUSCIMA_ROOT = _get_cvc_muscima_root()


##############################################################################


def _get_mff_muscima_root() -> Optional[str]:
    if 'MUSCIMA_PLUSPLUS_ROOT' in os.environ:
        muscima_plusplus_root = os.environ['MUSCIMA_PLUSPLUS_ROOT']
        return muscima_plusplus_root
    else:
        logging.info('muscima.dataset: environmental variable MUSCIMA_PLUSPLUS_ROOT not defined.')
        return None


MUSCIMA_PLUSPLUS_ROOT = _get_mff_muscima_root()


##############################################################################


class CvcMuscimaDataset(object):
    """The :class:`CvcMuscimaDataset` class implements a wrapper around
    the CVC-MUSCIMA dataset file structure that allows easy retrieval
    of filenames based on the page number (1 - 20), writer number
    (1 - 50), distortion, and mode (full image, staffline pixels only, or
    non-staffline pixels only).

    This functionality is defined in :meth:`imfile`.
    """

    DISTORTIONS = [
        'curvature',
        'ideal',
        'interrupted',
        'kanungo',
        'rotated',
        'staffline-thickness-variation-v1',
        'staffline-thickness-variation-v2',
        'staffline-y-variation-v1',
        'staffline-y-variation-v2',
        'thickness-ratio',
        'typeset-emulation',
        'whitespeckles',
    ]

    MODES = ['full', 'symbol', 'staff_only']

    def __init__(self, root: str = _get_cvc_muscima_root(), validate: bool = False):
        """The dataset is instantiated by providing the path to its root
        directory. If the ``CVC_MUSCIMA_ROOT`` variable is set, you do not
        have to provide anything."""
        if root is None:
            raise ValueError('CVC_MUSCIMA class needs the dataset root to serve'
                             ' any useful purpose: either set the CVC_MUSCIMA_ROOT'
                             ' environmental variable, or supply the root manually.')
        if not os.path.isdir(root):
            logging.warning('Instantiating CVC-MUSCIMA dataset wrapper without'
                            ' a valid root: the path {0} does not lead to'
                            ' a directory.'.format(root))

        if validate and not self.validate(fail_early=True):
            raise ValueError('CVC_MUSCIMA innstance with root {0} does not'
                             ' represent a valid CVC-MUSCIMA dataset copy.'
                             ''.format(self.root))

        self.root = root

    def imfile(self, page: int, writer: int, distortion: str = 'ideal', mode: str = 'full') -> str:
        """Construct the path leading to the file of the CVC-MUSCIMA image
        with the specified page (1 - 20), writer (1 - 50), distortion
        (see ``CVC_MUSCIMA_DISTORTIONS``), and mode (``full``, ``symbol``,
        ``staff_only``).

        This is the primary interface that the CVC_MUSCIMA class provides.
        """
        if distortion not in self.DISTORTIONS:
            raise ValueError('Distortion {0} not available in CVC-MUSCIMA.'
                             ' Choose one of:\n{1}'
                             ''.format(distortion, self.DISTORTIONS))
        if mode not in self.MODES:
            raise ValueError('Image mode {0} not available in CVC-MUSCIMA.'
                             ' Choose one of:\n{1}'
                             ''.format(mode, self.MODES))

        filename = os.path.join(self.root, distortion,
                                self.__number2writer_dir(writer),
                                self.__mode2dir(mode),
                                self.__number2page_file(page))
        if not os.path.isfile(filename):
            logging.warning('The requested file {0} should be available,'
                            ' but does not seem to be there. Are you sure'
                            ' the CVC-MUSCIMA root is set correctly? ({1})'
                            ''.format(filename, self.root))
        return filename

    def __number2page_file(self, n: int) -> str:
        if (n < 1) or (n > 20):
            raise ValueError('Invalid CVC-MUSCIMA score number {0}.'
                             ' Valid only between 1 and 20.'.format(n))
        if n < 10:
            return 'p0' + '0' + str(n) + '.png'
        else:
            return 'p0' + str(n) + '.png'

    def __number2writer_dir(self, n: int) -> str:
        if (n < 1) or (n > 50):
            raise ValueError('Invalid MUSCIMA writer number {0}.'
                             ' Valid only between 1 and 50.'.format(n))
        if n < 10:
            return 'w-0' + str(n)
        else:
            return 'w-' + str(n)

    def __mode2dir(self, mode: str) -> str:
        if mode == 'full':
            return 'image'
        elif mode == 'symbol':
            return 'symbol'
        elif mode == 'staff_only':
            return 'gt'

    def validate(self, fail_early: bool = True):
        """Checks whether the instantiated CVC_MUSCIMA instance really
        corresponds to the CVC-MUSCIMA dataset: all the 12 x 1000 expected
        CVC-MUSCIMA files should be present.

        :param fail_early: If ``True``, will return as soon as it encounters
            a missing file, if ``False``, will keep going through all the files
            and find out which ones are missing. (Default: ``True``)

        :returns: ``True`` if the dataset is OK, ``False`` if any file
            is missing.
        """
        _missing = []
        for d in self.DISTORTIONS:
            for m in self.MODES:
                for p in range(1, 21):
                    for w in range(1, 51):
                        f = self.imfile(page=p, writer=w, distortion=d, mode=m)
                        if not os.path.isfile(f):
                            if fail_early:
                                logging.warning('Missing file: {0}'.format(f))
                                return False
                            _missing.append(f)
        if len(_missing) > 0:
            logging.warning('Missing files in CVC_MUSCIMA instance with root {0}:'
                            '\n{1}'.format(self.root, '\n'.join(_missing)))
            return False
        return True



"""This is a script that removes all staffline, staffspace and staff
symbols, and all relationships that lead to them."""
import argparse
import copy
import logging
import os
import time

from mung.io import read_nodes_from_file, export_node_list
from mung.constants import InferenceEngineConstants as _CONST


def build_argument_parser():
    parser = argparse.ArgumentParser(description=__doc__, add_help=True,
                                     formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('-a', '--annot', action='store', required=True,
                        help='The annotation file for which the staffline and staff'
                             ' Node relationships should be added.')
    parser.add_argument('-e', '--export', action='store',
                        help='A filename to which the output NodeList'
                             ' should be saved. If not given, will print to'
                             ' stdout.')

    parser.add_argument('-v', '--verbose', action='store_true',
                        help='Turn on INFO messages.')
    parser.add_argument('--debug', action='store_true',
                        help='Turn on DEBUG messages.')

    return parser


def main(args):
    logging.info('Starting main...')
    _start_time = time.clock()

    # Your code goes here
    ##########################################################################
    logging.info('Import the Node list')
    if not os.path.isfile(args.annot):
        raise ValueError('Annotation file {0} not found!'
                         ''.format(args.annot))
    nodes = read_nodes_from_file(args.annot)

    ##########################################################################
    staff_id_to_node_mapping = {node.id: node for node in nodes
                              if node.class_name in _CONST.STAFF_CLASS_NAMES}

    output_nodes = []
    for node in nodes:
        if node.id in staff_id_to_node_mapping:
            continue
        new_c = copy.deepcopy(node)
        new_c.inlinks = [i for i in node.inlinks
                         if i not in staff_id_to_node_mapping]
        new_c.outlinks = [o for o in node.outlinks
                          if o not in staff_id_to_node_mapping]
        output_nodes.append(new_c)

    ##########################################################################
    logging.info('Export the stripped list.')
    nodes_string = export_node_list(output_nodes)

    if args.export is not None:
        with open(args.export, 'w') as hdl:
            hdl.write(nodes_string)
    else:
        print(nodes_string)

    _end_time = time.clock()
    logging.info('strip_staffline_symbols.py done in {0:.3f} s'.format(_end_time - _start_time))


if __name__ == '__main__':
    parser = build_argument_parser()
    args = parser.parse_args()

    if args.verbose:
        logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.INFO)
    if args.debug:
        logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.DEBUG)

    main(args)



"""This is a script that processes a set of symbols in order to obtain the pitch
recognition baseline. Intended to be used on top of an object detection stage."""
import argparse
import collections
import logging
import os
import pickle
import pprint
import time
import traceback

import numpy
from sklearn.feature_extraction import DictVectorizer
from typing import List, Dict, Tuple

from mung.grammar import DependencyGrammar
from mung.graph import find_beams_incoherent_with_stems, NotationGraph
from mung.graph import find_contained_nodes, remove_contained_nodes
from mung.graph import find_misdirected_leger_line_edges
from mung2midi.inference import OnsetsInferenceEngine, MIDIBuilder, PitchInferenceEngine
from mung.constants import InferenceEngineConstants as _CONST
from mung.io import parse_node_classes, read_nodes_from_file, export_node_list
from mung.node import bounding_box_intersection, merge_multiple_nodes, link_nodes, Node
from mung.stafflines import merge_staffline_segments, build_staff_nodes, \
    build_staffspace_nodes, \
    add_staff_relationships


def add_key_signatures(nodes: List[Node]) -> List[Node]:
    """Heuristic for deciding which accidentals are inline,
    and which should be interpreted as components of a key signature.

    Assumes staffline relationships have already been inferred.

    The heuristic is defined for each staff S as the following:

    * Take the leftmost clef C.
    * Take the leftmost notehead (incl. grace notes) N.
    * Take all accidentals A_S that overlap the space between C and N
      horizontally (including C, not including N), and overlap the staff.
    * Order A_S left-to-right
    * Set m = C
    * Initialize key signature K_S = {}
    * For each a_S in A_S:
    * if it is closer to m than to N, then:
    *   add a_S to K_S,
    *   set m = a_S
    * else:
    *   break

    Note that this modifies the accidentals and staffs in-place: they get inlinks
    from their respective key signatures.
    """
    graph = NotationGraph(nodes)

    new_node_id = max([m.id for m in nodes]) + 1

    key_signatures = []

    staffs = [m for m in nodes if m.class_name == _CONST.STAFF_CLASS_NAME]
    for s in staffs:

        # Take the leftmost clef C.
        clefs = graph.parents(s.id, class_filter=_CONST.CLEF_CLASS_NAMES)
        if len(clefs) == 0:
            continue
        leftmost_clef = min(clefs, key=lambda x: x.left)

        # Take the leftmost notehead (incl. grace notes) N.
        noteheads = graph.parents(s.id, class_filter=_CONST.NOTEHEAD_CLASS_NAMES)
        if len(noteheads) == 0:
            continue
        leftmost_notehead = min(noteheads, key=lambda x: x.left)

        # Take all accidentals A_S that fall between C and N
        # horizontally, and overlap the staff.
        all_accidentals = [m for m in nodes
                           if m.class_name in _CONST.ACCIDENTAL_CLASS_NAMES]
        relevant_acc_bbox = s.top, leftmost_clef.left, s.bottom, leftmost_notehead.left
        relevant_accidentals = [m for m in all_accidentals
                                if bounding_box_intersection(m.bounding_box,
                                                             relevant_acc_bbox)
                                is not None]

        # Order A_S left-to-right.
        ordered_accidentals = sorted(relevant_accidentals,
                                     key=lambda x: x.left)

        # Set m = C
        current_lstop = leftmost_clef.right
        key_signature_accidentals = []

        # Iterate over accidentals; check if they are closer to lstop
        # than to the first notehead
        for a in ordered_accidentals:
            if (a.left - current_lstop) < (leftmost_notehead.left - a.right):
                key_signature_accidentals.append(a)
                current_lstop = a.right
            else:
                break

        # Build key signature and connect it to staff
        if len(key_signature_accidentals) > 0:
            key_signature_class_names = list(_CONST.KEY_SIGNATURE_CLASS_NAMES)[0]
            # Note: there might be spurious links from the accidentals
            # to notheads, if they were mis-interpreted during parsing.
            # This actually might not matter; needs testing.
            key_signature = merge_multiple_nodes(
                key_signature_accidentals,
                class_name=key_signature_class_names,
                id_=new_node_id)
            new_node_id += 1
            link_nodes(key_signature, s)
            for a in key_signature_accidentals:
                link_nodes(key_signature, a)

            key_signatures.append(key_signature)

    logging.info('Adding {} key signatures'.format(len(key_signatures)))
    return nodes + key_signatures


##############################################################################
# Feature extraction

class PairwiseClassificationFeatureExtractor(object):
    def __init__(self, vectorizer=None):
        """Initialize the feature extractor.

        :param vectorizer: A DictVectorizer() from scikit-learn.
            Used to convert feature dicts to the vectors that
            the edge classifier of the parser will expect.
            If None, will create a new DictVectorizer. (This is useful
            for training; you can then pickle the entire extractor
            and make sure the feature extraction works for the classifier
            at runtime.)
        """
        if vectorizer is None:
            vectorizer = DictVectorizer()
        self.vectorizer = vectorizer

    def __call__(self, *args, **kwargs):
        """The call is per item (in this case, Node pair)."""
        fd = self.get_features_relative_bbox_and_clsname(*args, **kwargs)
        # Compensate for the vecotrizer "target", which we don't have here (by :-1)
        item_features = self.vectorizer.transform(fd).toarray()[0, :-1]
        return item_features

    def get_features_relative_bbox_and_clsname(self, c_from: Node, c_to: Node):
        """Extract a feature vector from the given pair of Nodes.
        Does *NOT* convert the class names to integers.

        Features: bbox(c_to) - bbox(c_from), class_name(c_from), class_name(c_to)
        Target: 1 if there is a link from u to v

        Returns a dict that works as input to ``self.vectorizer``.
        """
        target = 0
        if c_from.document == c_to.document:
            if c_to.id in c_from.outlinks:
                target = 1
        features = (c_to.top - c_from.top,
                    c_to.left - c_from.left,
                    c_to.bottom - c_from.bottom,
                    c_to.right - c_from.right,
                    c_from.class_name,
                    c_to.class_name,
                    target)
        dt, dl, db, dr, cu, cv, tgt = features
        # Normalizing clsnames
        if cu.startswith('letter'):
            cu = 'letter'
        if cu.startswith('numeral'):
            cu = 'numeral'
        if cv.startswith('letter'):
            cv = 'letter'
        if cv.startswith('numeral'):
            cv = 'numeral'
        feature_dict = {'dt': dt,
                        'dl': dl,
                        'db': db,
                        'dr': dr,
                        'cls_from': cu,
                        'cls_to': cv,
                        'target': tgt}
        return feature_dict

    def get_features_distance_relative_bbox_and_clsname(self, from_node: Node,
                                                        to_node: Node) -> Dict:
        """Extract a feature vector from the given pair of Nodes.
        Does *NOT* convert the class names to integers.

        Features: bbox(c_to) - bbox(c_from), class_name(c_from), class_name(c_to)
        Target: 1 if there is a link from u to v

        Returns a tuple.
        """
        target = 0
        if from_node.document == to_node.document:
            if to_node.id in from_node.outlinks:
                target = 1
        distance = from_node.distance_to(to_node)
        features = (distance,
                    to_node.top - from_node.top,
                    to_node.left - from_node.left,
                    to_node.bottom - from_node.bottom,
                    to_node.right - from_node.right,
                    from_node.class_name,
                    to_node.class_name,
                    target)
        dist, dt, dl, db, dr, cu, cv, tgt = features
        if cu.startswith('letter'):
            cu = 'letter'
        if cu.startswith('numeral'):
            cu = 'numeral'
        if cv.startswith('letter'):
            cv = 'letter'
        if cv.startswith('numeral'):
            cv = 'numeral'
        feature_dict = {'dist': dist,
                        'dt': dt,
                        'dl': dl,
                        'db': db,
                        'dr': dr,
                        'cls_from': cu,
                        'cls_to': cv,
                        'target': tgt}
        return feature_dict


##############################################################################
# Edge classifier


class PairwiseClassificationParser(object):
    """This parser applies a simple classifier that takes the bounding
    boxes of two Nodes and their classes and returns whether there
    is an edge or not."""
    MAXIMUM_DISTANCE_THRESHOLD = 200

    def __init__(self, grammar: DependencyGrammar, classifier,
                 feature_extractor: PairwiseClassificationFeatureExtractor):
        self.grammar = grammar
        self.classifier = classifier
        self.extractor = feature_extractor

    def parse(self, nodes: List[Node]):

        # Ensure the same docname for all Nodes,
        # since we later compute their distances.
        # The correct docname gets set on export anyway.
        pairs, features = self.extract_all_pairs(nodes)

        logging.info(
            'Clf.Parse: {0} object pairs from {1} objects'.format(len(pairs), len(nodes)))

        preds = self.classifier.predict(features)

        edges = []
        for idx, (c_from, c_to) in enumerate(pairs):
            if preds[idx] != 0:
                edges.append((c_from.id, c_to.id))

        edges = self.__apply_trivial_fixes(nodes, edges)
        return edges

    def __apply_trivial_fixes(self, nodes: List[Node], edges: List[Tuple[int, int]]):
        edges = self.__only_one_stem_per_notehead(nodes, edges)
        edges = self.__every_full_notehead_has_a_stem(nodes, edges)

        return edges

    def __only_one_stem_per_notehead(self, nodes: List[Node], edges: List[Tuple[int, int]]):
        node_id_to_node_mapping = {n.id: n for n in nodes}  # type: Dict[int, Node]

        # Collect stems per notehead
        stems_per_notehead = collections.defaultdict(list)
        stem_objids = set()
        for from_id, to_id in edges:
            from_node = node_id_to_node_mapping[from_id]
            to_node = node_id_to_node_mapping[to_id]
            if (from_node.class_name in _CONST.NOTEHEAD_CLASS_NAMES) and \
                    (to_node.class_name == 'stem'):
                stems_per_notehead[from_id].append(to_id)
                stem_objids.add(to_id)

        # Pick the closest one (by minimum distance)
        closest_stems_per_notehead = dict()
        for n_objid in stems_per_notehead:
            n = node_id_to_node_mapping[n_objid]
            stems = [node_id_to_node_mapping[objid] for objid in stems_per_notehead[n_objid]]
            closest_stem = min(stems, key=lambda s: n.distance_to(s))
            closest_stems_per_notehead[n_objid] = closest_stem.id

        # Filter the edges
        edges = [(f_objid, t_objid) for f_objid, t_objid in edges
                 if (f_objid not in closest_stems_per_notehead) or
                 (t_objid not in stem_objids) or
                 (closest_stems_per_notehead[f_objid] == t_objid)]

        return edges

    def __every_full_notehead_has_a_stem(self, nodes: List[Node], edges):
        node_id_to_node_mapping = {c.id: c for c in nodes}  # type: Dict[int, Node]

        # Collect stems per notehead
        notehead_objids = set([c.id for c in nodes if c.class_name == 'noteheadFull'])
        stem_objids = set([c.id for c in nodes if c.class_name == 'stem'])

        noteheads_with_stem_objids = set()
        stems_with_notehead_objids = set()
        for f, t in edges:
            if node_id_to_node_mapping[f].class_name == 'noteheadFull':
                if node_id_to_node_mapping[t].class_name == 'stem':
                    noteheads_with_stem_objids.add(f)
                    stems_with_notehead_objids.add(t)

        noteheads_without_stems = {n: node_id_to_node_mapping[n] for n in notehead_objids
                                   if n not in noteheads_with_stem_objids}
        stems_without_noteheads = {n: node_id_to_node_mapping[n] for n in stem_objids
                                   if n not in stems_with_notehead_objids}

        # To each notehead, assign the closest stem that is not yet taken.
        closest_stem_per_notehead = {objid: min(stems_without_noteheads,
                                                key=lambda x: node_id_to_node_mapping[
                                                    x].distance_to(n))
                                     for objid, n in list(noteheads_without_stems.items())}

        # Filter edges that are too long
        _n_before_filter = len(closest_stem_per_notehead)
        closest_stem_threshold_distance = 80
        closest_stem_per_notehead = {n_objid: s_objid
                                     for n_objid, s_objid in list(closest_stem_per_notehead.items())
                                     if node_id_to_node_mapping[n_objid].distance_to(
                node_id_to_node_mapping[s_objid]) < closest_stem_threshold_distance}

        return edges + list(closest_stem_per_notehead.items())

    def extract_all_pairs(self, nodes: List[Node]):
        pairs = []
        features = []
        for u in nodes:
            for v in nodes:
                if u.id == v.id:
                    continue
                distance = u.distance_to(v)
                if distance < self.MAXIMUM_DISTANCE_THRESHOLD:
                    pairs.append((u, v))
                    f = self.extractor(u, v)
                    features.append(f)

        # logging.info('Parsing features: {0}'.format(features[0]))
        features = numpy.array(features)
        # logging.info('Parsing features: {0}/{1}'.format(features.shape, features))
        return pairs, features

    def is_edge(self, c_from, c_to) -> bool:
        features = self.extractor(c_from, c_to)
        result = self.classifier.predict(features)
        return result

    def set_grammar(self, grammar: DependencyGrammar):
        self.grammar = grammar


def do_parse(nodes: List[Node], parser: PairwiseClassificationParser) -> List[Node]:
    non_staff_nodes = [node for node in nodes if node.class_name not in _CONST.STAFF_CLASS_NAMES]
    edges = parser.parse(non_staff_nodes)

    # Add edges
    id_to_node_mapping = {node.id: node for node in nodes}
    for f, t in edges:
        cf, ct = id_to_node_mapping[f], id_to_node_mapping[t]
        if t not in cf.outlinks:
            if f not in ct.inlinks:
                cf.outlinks.append(t)
                ct.inlinks.append(f)

    return nodes


##############################################################################
# Staffline building

def process_stafflines(nodes: List[Node],
                       do_build_staffs: bool = True,
                       do_build_staffspaces: bool = True,
                       do_add_staff_relationships: bool = True) -> List[Node]:
    """Merges staffline fragments into stafflines. Can group them into staffs,
    add staffspaces, and add the various obligatory relationships of other
    objects to the staff objects. Required before attempting to export MIDI."""
    if len([c for c in nodes if c.class_name == 'staff']) > 0:
        logging.warning('Some stafflines have already been processed. Reprocessing'
                        ' is not certain to work.')

    try:
        new_nodes = merge_staffline_segments(nodes)
    except ValueError as e:
        logging.warning('Model: Staffline merge failed:\n\t\t'
                        '{0}'.format(e.message))
        raise

    try:
        if do_build_staffs:
            staffs = build_staff_nodes(new_nodes)
            new_nodes = new_nodes + staffs
    except Exception as e:
        logging.warning('Building staffline Nodes from merged segments failed:'
                        ' {0}'.format(e.message))
        raise

    try:
        if do_build_staffspaces:
            staffspaces = build_staffspace_nodes(new_nodes)
            new_nodes = new_nodes + staffspaces
    except Exception as e:
        logging.warning('Building staffspace Nodes from stafflines failed:'
                        ' {0}'.format(e.message))
        raise

    try:
        if do_add_staff_relationships:
            new_nodes = add_staff_relationships(new_nodes)
    except Exception as e:
        logging.warning('Adding staff relationships failed:'
                        ' {0}'.format(e.message))
        raise

    return new_nodes


def find_wrong_edges(nodes: List[Node], grammar: DependencyGrammar) -> List[Tuple[int, int]]:
    id_to_node_mapping = {node.id: node for node in nodes}
    graph = NotationGraph(nodes)

    incoherent_beam_pairs = find_beams_incoherent_with_stems(nodes)
    # Switching off misdirected leger lines: there is something wrong with them
    misdirected_leger_lines = find_misdirected_leger_line_edges(nodes)

    wrong_edges = [(n.id, b.id)
                   for n, b in incoherent_beam_pairs + misdirected_leger_lines]

    disallowed_symbol_class_pairs = [(f, t) for f, t in graph.edges
                                     if not grammar.validate_edge(id_to_node_mapping[f].class_name,
                                                                  id_to_node_mapping[t].class_name)]
    wrong_edges += disallowed_symbol_class_pairs
    return wrong_edges


def find_very_small_nodes(nodes: List[Node], bbox_threshold=40, mask_threshold=35) -> List[int]:
    very_small_nodes = []

    for c in nodes:
        total_masked_area = c.mask.sum()
        total_bbox_area = c.width * c.height
        if total_bbox_area < bbox_threshold:
            very_small_nodes.append(c)
        elif total_masked_area < mask_threshold:
            very_small_nodes.append(c)

    return list(set([c.id for c in very_small_nodes]))


def infer_precedence_edges(nodes: List[Node], factor_by_staff=True) -> List[Tuple[int, int]]:
    """Returns a list of (from_objid, to_objid) parirs. They
    then need to be added to the Nodes as precedence edges."""
    id_to_node_mapping = {c.id: c for c in nodes}
    relevant_class_names = set(list(_CONST.NONGRACE_NOTEHEAD_CLASS_NAMES)
                               + list(_CONST.REST_CLASS_NAMES))
    precedence_nodes = [c for c in nodes
                        if c.class_name in relevant_class_names]
    logging.info('_infer_precedence: {0} total prec. Nodes'
                 ''.format(len(precedence_nodes)))

    # Group the objects according to the staff they are related to
    # and infer precedence on these subgroups.
    if factor_by_staff:
        staffs = [c for c in nodes
                  if c.class_name == _CONST.STAFF_CLASS_NAME]
        logging.info('_infer_precedence: got {0} staffs'.format(len(staffs)))
        staff_objids = {c.id: i for i, c in enumerate(staffs)}
        precedence_nodes_per_staff = [[] for _ in staffs]
        # All Nodes relevant for precedence have a relationship
        # to a staff.
        for c in precedence_nodes:
            for o in c.outlinks:
                if o in staff_objids:
                    precedence_nodes_per_staff[staff_objids[o]].append(c)

        logging.info('Precedence groups: {0}'
                     ''.format(precedence_nodes_per_staff))
        prec_edges = []
        for precedence_nodes_group in precedence_nodes_per_staff:
            group_prec_edges = infer_precedence_edges(precedence_nodes_group,
                                                      factor_by_staff=False)
            prec_edges.extend(group_prec_edges)
        return prec_edges

    if len(precedence_nodes) <= 1:
        logging.info('EdgeListView._infer_precedence: less than 2'
                     ' timed Nodes selected, no precedence'
                     ' edges to infer.')
        return []

    # Group into equivalence if noteheads share stems
    _stems_to_noteheads_map = collections.defaultdict(list)
    for c in precedence_nodes:
        for o in c.outlinks:
            if o not in id_to_node_mapping:
                logging.warning('Dangling outlink: {} --> {}'.format(c.id, o))
                continue
            c_o = id_to_node_mapping[o]
            if c_o.class_name == 'stem':
                _stems_to_noteheads_map[c_o.id].append(c.id)

    _prec_equiv_objids = []
    _stemmed_noteheads_objids = []
    for _stem_objid, _stem_notehead_objids in list(_stems_to_noteheads_map.items()):
        _stemmed_noteheads_objids = _stemmed_noteheads_objids \
                                    + _stem_notehead_objids
        _prec_equiv_objids.append(_stem_notehead_objids)
    for c in precedence_nodes:
        if c.id not in _stemmed_noteheads_objids:
            _prec_equiv_objids.append([c.id])

    equiv_objs = [[id_to_node_mapping[objid] for objid in equiv_objids]
                  for equiv_objids in _prec_equiv_objids]

    # Order the equivalence classes left to right
    sorted_equiv_objs = sorted(equiv_objs,
                               key=lambda eo: min([o.left for o in eo]))

    edges = []
    for i in range(len(sorted_equiv_objs) - 1):
        fr_objs = sorted_equiv_objs[i]
        to_objs = sorted_equiv_objs[i + 1]
        for f in fr_objs:
            for t in to_objs:
                edges.append((f.id, t.id))

    return edges


def add_precedence_edges(nodes: List[Node], edges: List[Tuple[int, int]]) -> List[Node]:
    """Adds precedence edges to Nodes."""
    # Ensure unique
    edges = set(edges)
    id_to_node_mapping = {c.id: c for c in nodes}

    for f, t in edges:
        cf, ct = id_to_node_mapping[f], id_to_node_mapping[t]

        if cf.data is None:
            cf.data = dict()
        if 'precedence_outlinks' not in cf.data:
            cf.data['precedence_outlinks'] = []
        cf.data['precedence_outlinks'].append(t)

        if ct.data is None:
            ct.data = dict()
        if 'precedence_inlinks' not in ct.data:
            ct.data['precedence_inlinks'] = []
        ct.data['precedence_inlinks'].append(f)

    return nodes


def build_midi(nodes: List[Node], selected_nodes: List[Node] = None,
               retain_pitches: bool = True,
               retain_durations: bool = True,
               retain_onsets: bool = True,
               tempo: int = 180):
    """Attempts to export a MIDI file from the current graph. Assumes that
    all the staff objects and their relations have been correctly established,
    and that the correct precedence graph is available.

    :param retain_pitches: If set, will record the pitch information
        in pitched objects.

    :param retain_durations: If set, will record the duration information
        in objects to which it applies.

    :returns: A single-track ``midiutil.MidiFile.MIDIFile`` object. It can be
        written to a stream using its ``mf.writeFile()`` method."""
    id_to_node_mapping = {c.id: c for c in nodes}

    pitch_inference_engine = PitchInferenceEngine()
    time_inference_engine = OnsetsInferenceEngine(nodes=nodes)

    try:
        logging.info('Running pitch inference.')
        pitches, pitch_names = pitch_inference_engine.infer_pitches(nodes,
                                                                    with_names=True)
    except Exception as e:
        logging.warning('Model: Pitch inference failed!')
        logging.exception(traceback.format_exc(e))
        raise

    if retain_pitches:
        for objid in pitches:
            c = id_to_node_mapping[objid]
            pitch_step, pitch_octave = pitch_names[objid]
            c.data['midi_pitch_code'] = pitches[objid]
            c.data['normalized_pitch_step'] = pitch_step
            c.data['pitch_octave'] = pitch_octave

    try:
        logging.info('Running durations inference.')
        durations = time_inference_engine.durations(nodes)
    except Exception as e:
        logging.warning('Model: Duration inference failed!')
        logging.exception(traceback.format_exc(e))
        raise

    if retain_durations:
        for objid in durations:
            c = id_to_node_mapping[objid]
            c.data['duration_beats'] = durations[objid]

    try:
        logging.info('Running onsets inference.')
        onsets = time_inference_engine.onsets(nodes)
    except Exception as e:
        logging.warning('Model: Onset inference failed!')
        logging.exception(traceback.format_exc(e))
        raise

    if retain_onsets:
        for objid in onsets:
            c = id_to_node_mapping[objid]
            c.data['onset_beats'] = onsets[objid]

    # Process ties
    durations, onsets = time_inference_engine.process_ties(nodes, durations, onsets)

    # Prepare selection subset
    if selected_nodes is None:
        selected_nodes = nodes
    ids_of_selected_nodes = [c.id for c in selected_nodes]

    # Build the MIDI data
    midi_builder = MIDIBuilder()
    mf = midi_builder.build_midi(
        pitches=pitches, durations=durations, onsets=onsets,
        selection=ids_of_selected_nodes, tempo=tempo)

    return mf


def build_argument_parser():
    parser = argparse.ArgumentParser(description=__doc__, add_help=True,
                                     formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('-i', '--input_mung', action='store', required=True,
                        help='Read the input MuNG to this'
                             ' file.')
    parser.add_argument('-o', '--output_mung', action='store', required=True,
                        help='Write the resulting MuNG to this'
                             ' file.')

    parser.add_argument('--mlclasses', action='store', required=True,
                        help='Read the NodeClass list from this XML.')
    parser.add_argument('--grammar', action='store', required=True,
                        help='Read the grammar file that specifies the allowed edge node'
                             ' class pairs.')
    parser.add_argument('--parser', action='store', required=True,
                        help='Read the pickled feature extractor for parser classifier.')
    parser.add_argument('--vectorizer', action='store', required=True,
                        help='Read the pickled parser classifier.')

    parser.add_argument('--add_key_signatures', action='store_true',
                        help='Attempt to add key signatures. Algorithm is'
                             ' very basic: for each staff, gather all accidentals'
                             ' intersecting that staff. From the left, find the'
                             ' first notehead on a staff that is not contained in'
                             ' anything, and the last clef before'
                             ' this notehead. (...) [TESTING]')
    parser.add_argument('--filter_contained', action='store_true',
                        help='Remove objects that are fully contained within another.'
                             ' This should be applied *after* parsing non-staff'
                             ' edges, but *before* inferring staffs.')

    parser.add_argument('-v', '--verbose', action='store_true',
                        help='Turn on INFO messages.')
    parser.add_argument('--debug', action='store_true',
                        help='Turn on DEBUG messages.')

    return parser


def main(args):
    logging.info('Starting main...')
    _start_time = time.clock()

    ###############################################################
    # Preparation: loading the parsing apparatus

    with open(args.vectorizer) as hdl:
        vectorizer = pickle.load(hdl)
    feature_extractor = PairwiseClassificationFeatureExtractor(vectorizer=vectorizer)

    with open(args.parser) as hdl:
        classifier = pickle.load(hdl)

    mlclass_list = parse_node_classes(args.mlclasses)
    mlclasses = {m.name for m in mlclass_list}

    grammar = DependencyGrammar(grammar_filename=args.grammar, alphabet=mlclasses)

    parser = PairwiseClassificationParser(grammar=grammar,
                                          classifier=classifier,
                                          feature_extractor=feature_extractor)

    #################################################################
    logging.info('Load graph')
    nodes = read_nodes_from_file(args.input_mung)

    logging.info('Filter very small')
    very_small_nodes = find_very_small_nodes(nodes,
                                             bbox_threshold=40,
                                             mask_threshold=35)
    very_small_nodes = set(very_small_nodes)
    nodes = [c for c in nodes if c not in very_small_nodes]

    logging.info('Parsing')
    nodes = do_parse(nodes, parser=parser)

    # Filter contained here.
    if args.filter_contained:
        logging.info('Finding contained Nodes...')
        contained = find_contained_nodes(nodes,
                                         mask_threshold=0.95)
        NEVER_DISCARD_CLASSES = ['key_signature', 'time_signature']
        contained = [c for c in contained if c.class_name not in NEVER_DISCARD_CLASSES]

        _contained_counts = collections.defaultdict(int)
        for c in contained:
            _contained_counts[c.class_name] += 1
        logging.info('Found {} contained Nodes'.format(len(contained)))
        logging.info('Contained counts:\n{0}'.format(pprint.pformat(dict(_contained_counts))))
        nodes = remove_contained_nodes(nodes,
                                       contained)
        logging.info('Removed contained Nodes: {}...'.format([m.id for m in contained]))

    logging.info('Inferring staffline & staff objects, staff relationships')
    nodes = process_stafflines(nodes)

    if args.add_key_signatures:
        nodes = add_key_signatures(nodes)

    logging.info('Filter invalid edges')
    graph = NotationGraph(nodes)
    # Operatng on the graph changes the Nodes
    #  -- the graph only keeps a pointer
    wrong_edges = find_wrong_edges(nodes, grammar)
    for f, t in wrong_edges:
        graph.remove_edge(f, t)

    logging.info('Add precedence relationships, factored only by staff')
    prec_edges = infer_precedence_edges(nodes)
    nodes = add_precedence_edges(nodes, prec_edges)

    logging.info('Ensuring MIDI can be built')
    mf = build_midi(nodes,
                    retain_pitches=True,
                    retain_durations=True,
                    retain_onsets=True,
                    tempo=180)

    logging.info('Save output')
    docname = os.path.splitext(os.path.basename(args.output_mung))[0]
    xml = export_node_list(nodes,
                           document=docname,
                           dataset='FNOMR_results')
    with open(args.output_mung, 'w') as out_stream:
        out_stream.write(xml)
        out_stream.write('\n')

    _end_time = time.clock()
    logging.info('baseline_process_symbols.py done in {0:.3f} s'.format(_end_time - _start_time))


if __name__ == '__main__':
    parser = build_argument_parser()
    args = parser.parse_args()

    if args.verbose:
        logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.INFO)
    if args.debug:
        logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.DEBUG)

    main(args)
